#include <iostream>
#include "mm_colour_properties_tool.h"
#include <sys/timeb.h>
#include <cassert>
#include "mm_object.h"

using std::cerr;
using std::cout;
using std::endl;

using namespace mimas;

//#include "mm_triangularization.h"

/**************************************************************
 *@author Manuel Boissenin
 *@date february 2004
 *
 *Calculate the disparity map of two rectified images:
 *images where correspondent points are on the same line.
 *
 *SAD: Sum of Absolute Difference
 *
 *This local method is based on SAD correlation. It is using 
 *a sliding windows implementation.
 *
 *
 *We assume that the disparity is only positive 
 *the speed is independant of the correlation size window.
 *
 *memory size is only O(width * depth * window_height) which is
 *the size of the ring buffer.
 *
 *use register values to fasten the calculation.
 *
 *colour added for correlation and displayed image.
 *
 *memory checked with valgrind
 *
 *Right-left consistency check implemented. Inconsistencies appear in red on the disparity map
 *
 *Subpixel estimation implemented. Still remains some problems.
 *
 *
 *
 *
 *to be implemented:
 *-subpixel resolution.
 *-mean to avoid problem of illumination
 *-median filter
 *-uniqueness of the minimum (signaled in blue for instance) (left-right concistency done other methods exists) 
 *
 *for mimas, setPixel(int x,int y,char *color) with basic color 
 *(red green blue yellow orange black white, magenta, cyan ... ...)
 *
 *to avoid stair effect tolerance has been added while doing the right left
 *consistency. This can be improved for instance when the difference is not very high
 *the pixel can be set blue and after filtering (mean, median ...) the pixel can
 *be set according to its neighbours
 *
 *convention: when the "pb" "modif" "manu" words appear it is to signal a problem 
 **************************************************************/


//This should increase the overall speed saving into memory some calculation already done.
//nevertheless it slow down the process. Using valgind it should be interesting finding why.
//Hypothesis: cache misses (to check with cachegrind)
//#define SAV_ACTIVATED //faster if not defined



class mm_disparity_map{
	private:
		mm_image< mm_rgba<int> > &left;
		mm_image< mm_rgba<int> > &right;

		int width;
		int height;
		int depth;

		int window_width;
		int window_height;

		int w_hlp;
		int h_hlp;


#if defined SAV_ACTIVATED  
		unsigned int ***ringBuffer; //putting const??
#else
		unsigned int **ringBuffer; 
#endif

		unsigned int *corr;

		unsigned int **res;//to find occluding pixels doing left-right consistency checking

		int min(int x, int y)
		{
			if (x < y)
				return x;
			return y;
		}
		int abs(int x)
		{
			if(x < 0) x = -x;
			return x;
		}

		float abs(float x)
		{
			if(x < 0) x = -x;
			return x;
		}

		int getValue(int x,int y,int d)
		{
         const mm_rgba< int > *leftPixel = &left.getPixel( x, y );
         const mm_rgba< int > *rightPixel = &right.getPixel( x+d, y );
			return abs( leftPixel->getRed() - rightPixel->getRed())
				+ abs(leftPixel->getGreen() - rightPixel->getGreen())
				+ abs(leftPixel->getBlue() - rightPixel->getBlue());
		}


		//do an object for the ringbuffer?
		void  initializeRingBuffer(void)
		{

#if defined SAV_ACTIVATED
			ringBuffer = new unsigned int**[width];
#else
			ringBuffer = new unsigned int*[width];
#endif
			for (int x = 0;x < width; x++)
			{
				int depth_max = min(depth, width - x);
#if defined SAV_ACTIVATED
				ringBuffer[x] = new unsigned int*[depth_max];
#else
				ringBuffer[x] = new unsigned int[depth_max];
#endif

				//Fill the ringBuffer with initial values	
				for (int d = 0; d < depth_max; d++)
				{
#if defined SAV_ACTIVATED
					ringBuffer[x][d] = new unsigned int[window_height + 1];
					ringBuffer[x][d][window_height] = 0;
#else
					ringBuffer[x][d]= 0;
#endif

					for (int y = 0; y < window_height; y++)
					{
						int sad = getValue(x,y,d);
#if defined SAV_ACTIVATED
						ringBuffer[x][d][y] = sad;
						ringBuffer[x][d][window_height] += sad;
#else
						ringBuffer[x][d] += sad;
#endif

					} 	    
				}
			}
		}

		void shiftRingBuffer(int y)
		{
			for (int x = 0;x < width; x++)
			{
				int depth_max = min(depth, width - x);

				//Fill the ringBuffer with initial values	
				for (int d = 0; d < depth_max; d++)
				{
#if defined SAV_ACTIVATED
					int newSad = getValue(x, y + h_hlp + 1, d);
					int oldSad = ringBuffer[x][d][(y - h_hlp)%window_height];
					ringBuffer[x][d][(y + h_hlp + 1) %window_height] = newSad;//check if same place (it should)
					ringBuffer[x][d][window_height] +=newSad - oldSad;// - getValue(x, y - h_hlp, d) + newSad;
#else
					ringBuffer[x][d] +=  getValue(x, y + h_hlp + 1, d) - getValue(x, y - h_hlp, d);
#endif


				}
			}
		}


		void initializeXRingBuffer(int line)
		{
			unsigned int min = 1000000;
			int chosenDepth = 0;

			for (int d = 0; d < depth; d ++)
			{
				//initialize to 0
				corr[d] = 0;
				//calculate sad value
				for(int x = 0; x < window_width; x++)
				{
#if defined SAV_ACTIVATED
					corr[d] += ringBuffer[x][d][window_height];
#else
					corr[d] += ringBuffer[x][d];
#endif
				}
				//find min of sad value
				if(corr[d] < min)
				{
					min = corr[d];
					chosenDepth = d;
				}
			}
			assert(min != 1000000);
			disparity_map.setPixel( w_hlp, line, mm_rgba< float >( chosenDepth,
                                                                chosenDepth,
                                                                chosenDepth ) );
		}


		//y only used to know where to set the disparity map value
		//carefull with x bound, determined by the calling function. Some improvments can be made here.
		void findDisparity(int x, int y)
		{
			int max_depth = min(depth, width - x - w_hlp);
			//here big pb bug mod error check bounds and why it works in implementation main.cc.4
			unsigned int min = 1000000;
			float chosenDepth = 0;


			for (int d = 0; d < max_depth; d ++)
			{
				//calculate sad value

				//	cout<<"x + w_hlp: "<<x + w_hlp<<" x - w_hlp -1: "<<x - w_hlp -1<<"d: "<< d<<endl;

#if defined SAV_ACTIVATED
				corr[d] += ringBuffer[x + w_hlp][d][window_height] - ringBuffer[x - w_hlp -1][d][window_height];
#else
				corr[d] += ringBuffer[x + w_hlp][d] - ringBuffer[x - w_hlp -1][d];
#endif


				res[x][d] = corr[d];//save result for the right-left consistency checking
				//find min of sad value
				if(corr[d] < min)
				{
					min = corr[d];
					chosenDepth = d;
				}
			}
			//  assert(min != 100000);

			/***********************SUBPIXEL ESTIMATION ************************************/
			//here we do the subpixel estimation
			if (chosenDepth>0 && chosenDepth<max_depth)
			{
				float up = float(corr[int(chosenDepth)-1]) - float(corr[int(chosenDepth)+1]);
				float down = 2*(float(corr[int(chosenDepth)-1]) - 2*float(corr[int(chosenDepth)])
						+ float(corr[int(chosenDepth)+1]));
				float quotient = up/down;


				if(quotient <-1 || quotient >1)
				{
					//pb
					//cout<<"problem, quotient: "<< quotient<<endl;
					//We should not have these cases. check what happen here.
				} 
				else
					chosenDepth = chosenDepth + quotient;
				/*	cout<<"dessus"<<up <<endl;
					cout<<"dessous"<<down<<endl;
					cout<< "added" << quotient <<endl;*/
			}


			disparity_map.setPixel( x, y, mm_rgba< float >( chosenDepth,
                                                         chosenDepth,
                                                         chosenDepth ) );



			/*****************************RIGHT_LEFT CONSISTENCY******************************/
			//do the right-left consistency checking and display in red pixels that doesn't satisfy it.
			//ofteenly those pixels are considered as occluded.

			int RightChosenDepth = 0;
			int _x = x;
			int _d = 0;

			//the test condition probably can be improved
			for (unsigned int minimum = res[x][0];(_d < depth ) && (_x > w_hlp + 1)  ;_d++,_x--) 
			{
				if(minimum > res[_x][_d])
				{
					minimum = res[_x][_d];
					RightChosenDepth = _d;
				}
			}

			//to avoid stair effect between values add a tolerance error:
			float diff = abs(disparity_map.getPixel(x - RightChosenDepth, y).getRed() - RightChosenDepth);

			//    if(disparity_map.getPixelRed(x - RightChosenDepth, y) != RightChosenDepth)
			if(diff > 1) 
			{
           disparity_map.setPixel( x - RightChosenDepth, y,
                                   mm_rgba< float >( 255, 0, 0 ) );
			}


		}

	public:

		mm_image< mm_rgba< float > > disparity_map;//we need float because of subpixel estimation

		/*
		   ~mm_disparity_map(void)
		   {
		   delete []corr;

		   for (int x = 0;x < width; x++)
		   {
		   int depth_max = min(depth, width - x);
		   for (int d = 0; d < depth_max; d++)
		   {
		   delete[] ringBuffer[x][d];
		   }
		   }

		   for (int x = 0;x < width; x++)
		   {
		   delete[] ringBuffer[x];
		   }

		   delete []ringBuffer;
		   }

*/

		void fastCalculateDisparityMap(void)
		{
			initializeRingBuffer();
			for (int y = h_hlp; y < height - h_hlp; y++)
			{	
				initializeXRingBuffer(y);	
				for (int x = w_hlp + 1; x < width - w_hlp; x++)
					//initializeXRingBuffer already calculate value for w_hlp
				{
					findDisparity(x,y);
				}

				//change here
				if (y != height - h_hlp -1)
					shiftRingBuffer(y);
			}
		}

		//odd value are necessary for the window sizes
		mm_disparity_map( mm_image< mm_rgba< int > > &left,
                        mm_image< mm_rgba< int > > &right, int depth,
				int window_width = 7, int window_height = 7)
			:left(left),right(right),
		window_width(window_width),window_height(window_height),
		w_hlp((window_width-1)/2),h_hlp((window_height-1)/2)
		{
			//check size of images  are  the same 
			width = left.getWidth();
			height = right.getHeight();

			corr = new unsigned int[depth];//used to stock sad values of one position(x,y)

			res = new unsigned int*[width];

			for(int x = 0; x < width; x++)
			{
				int max_depth = min(depth, width - x);
				res[x] = new unsigned int[max_depth];
			} 

			this->depth = depth;
			disparity_map.init(width,height);//all value initialised at 0
		}
};
