#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <map>
#include "mm_exception.h"
#include "mm_mesa_offscreen.h"
#include "mm_mesa_onscreen.h"
#include "mm_timer.h"
#include "mm_image_mesaoutput.h"

using namespace boost;
using namespace std;
using namespace mimas;

mm_image< mm_rgba< unsigned char > > gaudy(void)
{
  mm_image< mm_rgba< unsigned char > > image; image.init( 256, 256 );
  for ( int k=0; k<=255; k++ )
    for ( int j=0; j<=255; j++ )
      image.pixel( k, j ) = mm_rgba< unsigned char >( k, j, 255 - k );
  return image;
}

void rotatedText( mm_mesa_base &m, int angle, const string &text ) {
  m.rectangle( 0, 0, 255, 255 );
  
  glPushMatrix();
  glTranslated( 128, 128, 0 );
  glScaled( 0.1, -0.1, 1.0 );
  glRotatef( angle, 0, 0, 1 );
  m.text( text );
  glPopMatrix();
}

class mimaswindow
{
 public:
  mimaswindow( mm_x11_display_ptr display ):
    image( gaudy() ),
    window( new mm_x11_window( display, 256, 256, &callback, this ) ) {}
  void update( int i ) {
    window->process_events();
    angle = i * 360 / 255;
    paint();
  };
 protected:
  static void callback( void *arg ) { ((mimaswindow *)arg)->paint(); }
  void paint(void) {
    mm_mesa_onscreen mesa( window );
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    glRasterPos2i( -1, 1 );
    glPixelZoom( 1, -1 );
    glDrawPixels( window->get_width(), window->get_height(),
                  GL_BGRA, GL_UNSIGNED_BYTE, image.rawData().data() );  
    mesa.ortho2D();
    const mm_rgba< unsigned char > red( 255, 0, 0 );
    mesa.setColor( red );
    rotatedText( mesa, angle, "onscreen" );
    mesa.finish();
  };
  mm_image< mm_rgba< unsigned char > > image;
  mm_x11_window_ptr window;
  int angle;
};

void paint( void *arg )
{
}

int main( int argc, char *argv[] )
{
  int retVal = 0;

  try {

    // Open X11 display.
    mm_x11_display_ptr display( new mm_x11_display );

    const int delay = 50;

    // First demo:
    // This demo demonstrates onscreen-rendering. This will consume less
    // processing time, especially on graphic-cards with 3D-rendering
    // capabilities.
    {
      mimaswindow window( display );
      for ( int i=0; i<255; i += 10 ) {
        mm_timer t;
        window.update( i );
        double remaining = delay * 0.001 - t.elapsed();
        if ( remaining > 0 ) usleep( (int)( remaining * 1.0E+6 ) );
      };
    };

    // Second demo:
    // Off-screen rendering on images.
    // Displaying images on two windows.
    // The second window is displaying the results of offscreen-rendering
    // converted to grayscale.
    {
      // Create two image-output windows.
      mm_image_mesaoutput< unsigned char > window1( display );
      mm_image_mesaoutput< mm_rgba< unsigned char > > window2( display );

      // Create some coloured image.
      mm_image< mm_rgba< unsigned char > > background( gaudy() );
      
      // Loop.
      for ( int i=0; i<=255; i += 10 ) {
        mm_image< unsigned char > image1( background );
        mm_image< mm_rgba< unsigned char > > image2( background );

        // Do some off-screen rendering on them.
        {
          mm_mesa_offscreen< unsigned char >
            m( display, image1, false );
          m.ortho2D();
          const mm_rgba< unsigned char > green( 0, 255, 0 );
          m.setColor( green );
          rotatedText( m, i * 360 / 255, "offscreen" );
        };

        {
          mm_mesa_offscreen< mm_rgba< unsigned char > >
            m( display, image2, false );
          m.ortho2D();
          const mm_rgba< unsigned char > green( 0, 255, 0 );
          m.setColor( green );
          // m.setGray( 255 );
          rotatedText( m, i * 360 / 255, "offscreen" );
        };

        // Update display.
        window1 << image1 << mimas::delay( delay );
        window2 << image2 << mimas::delay( delay );
      };
    };

  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  
  return retVal;
}
