#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <mm_correlate_square_region.h>
#include <algorithm>
#include <cassert>
#include <fstream>
#include <iomanip>
#include <iterator>
#include <map>
#include <mm_corner.h>
#include <mm_corner_tool.h>
#include <mm_corrcoef.h>
#include <mm_image.h>
#include <mm_image_fileinput.h>
#include <mm_mesa_onscreen.h>
#include <mm_svd_correspondence.h>
#include <sstream>
#include <vector>


#define SKIP 0
#define NUMIMAGES 10

using namespace std;
using namespace mimas;
using namespace boost;

typedef numeric::ublas::vector< float > FloatVector;
typedef numeric::ublas::vector< int > IntVector;

class CorrelationWeightedProximity:
  public binary_function< shared_ptr< IntVector >, shared_ptr< IntVector >,
                          double >
{
 public:
  CorrelationWeightedProximity( shared_ptr< mm_image< int > > aImage1,
                                shared_ptr< mm_image< int > > aImage2,
                                int aAreaDiv2,
                                double aSigma ):
    pixelCorrelation( aImage1, aImage2, aAreaDiv2, corrcoef ),
    sigma( aSigma ) {}
  double operator()( const shared_ptr< IntVector > &a,
                     const shared_ptr< IntVector > &b ) {
    double
      corr = pixelCorrelation( *a, *b ),
      dist = norm_2( FloatVector( *b - *a ) );
    return
      ( ( corr + 1 ) / 2.0 ) *
      ( ( corr + 1 ) / 2.0 ) *
      exp( -dist / sigma );
  }
 protected:
  mm_corrcoef< vector< int > > corrcoef;
  mm_correlate_square_region< int, mm_corrcoef< vector< int > > >
     pixelCorrelation;
  double sigma;
};

class flow_window
{
 public:
  flow_window( mm_x11_display_ptr display, int _width, int _height ):
    window( new mm_x11_window( display, _width, _height, &callback, this ) )
  { image.init( _width, _height ); }
  void update( const mm_image< int > &_image,
               shared_ptr< map< int, shared_ptr< IntVector > > > _tracking ) {
    image = mm_image< mm_rgba< unsigned char > >( _image );
    trackingOld = tracking;
    tracking = _tracking;
    window->process_events();
    paint();
  };
 protected:
  static void callback( void *arg ) { ((flow_window *)arg)->paint(); }
  void paint(void) {
    mm_mesa_onscreen mesa( window );
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    glRasterPos2i( -1, 1 );
    glPixelZoom( 1, -1 );
    glDrawPixels( window->get_width(), window->get_height(),
                  GL_BGRA, GL_UNSIGNED_BYTE, image.rawData().data() );  

    mesa.ortho2D();

    mm_rgba< unsigned char > yellow( 255, 255, 0 ), red( 255, 0, 0 );
    
    for ( map< int, shared_ptr< IntVector > >::const_iterator i =
            tracking->begin(); i != tracking->end(); i++ ) {
      
      shared_ptr< IntVector > vectorNew = i->second;
      
      int
        x = (int)(*vectorNew)( 0 ),
        y = (int)(*vectorNew)( 1 );
      
      glBegin( GL_LINES ); {
        
        if ( trackingOld ) {
          mesa.setColor( yellow );
          shared_ptr< IntVector > vector = (*trackingOld)[ i->first ];
          if ( vector ) {
            int
              xold = (int)(*vector)( 0 ),
              yold = (int)(*vector)( 1 );
            glVertex2i( xold, yold );
            glVertex2i( x   , y    );
          };
        };
        
        mesa.setColor( red );

        glVertex2i( x - 1, y );
        glVertex2i( x + 1, y );
        glVertex2i( x, y - 1 );
        glVertex2i( x, y + 1 );
          
      }; glEnd();
      
      ostringstream stream; stream << i->first;
      glPushMatrix(); {
        glTranslated( x + 4, y + 14, 0 );
        glScaled( 0.05, -0.05, 1.0 );
        mesa.text( stream.str() );
      }; glPopMatrix();

    };
      
    mesa.finish();

  };
  mm_image< mm_rgba< unsigned char > > image;
  shared_ptr< map< int, shared_ptr< IntVector > > > trackingOld, tracking;
  mm_x11_window_ptr window;
};

typedef shared_ptr< flow_window > flow_window_ptr;

int main( int argc, char *argv[] )
{
  int retVal = 0;
  try {
  MMERROR( argc >= 5, mm_exception, ,
           "Syntax : " << argv[0] << " <sigma> <area2> <corner threshold> "
           "<output filename prefix> files ..." << endl
           << "Example: " << argv[0] << " 60 3 0.9 result input*.pgm" );
  double sigma = atof( argv[ 1 ] );
  int area2 = atoi( argv[ 2 ] );
  double threshold = atof( argv[ 3 ] );

  mm_x11_display_ptr display( new mm_x11_display );
  flow_window_ptr window;

  shared_ptr< map< int, shared_ptr< IntVector > > > tracking;
  shared_ptr< mm_image< int > > image;

  int index = 0;
  
  for ( int c=0; c<argc-5; c++ ) {
	 
    shared_ptr< mm_image< int > > imageNew( new mm_image<int> );
    ifstream file( argv[ c+ 5 ], ios::binary ); file >> *imageNew;

    if ( !window )
      window = flow_window_ptr( new flow_window( display,
                                                 imageNew->getWidth(),
                                                 imageNew->getHeight() ) );
    
    mm_corner_image corners;

    mm_corner_tool<int> cornerTool( threshold );
    cornerTool.filterImage( *imageNew, corners );

    shared_ptr< vector< shared_ptr< IntVector > > > featuresNew
      ( new vector< shared_ptr< IntVector > > );

    for ( int i=0; i<(signed)corners.rawData().num_elements(); i++ ) {
      mm_corner_ptr corner = dynamic_pointer_cast< mm_corner >
        ( corners.rawData().data()[i] );
      if ( corner )
        featuresNew->
          push_back( shared_ptr< IntVector >
                     ( new IntVector( corner->getPosition() ) ) );
    };
      
    // Initialise map for tracking.
    shared_ptr< map< int, shared_ptr< IntVector > > > trackingNew
      ( new map< int, shared_ptr< IntVector > > );
    shared_ptr< vector< shared_ptr< IntVector > > > lostFeatures
      ( new vector< shared_ptr< IntVector > > );

    if ( tracking ) { 
      
      CorrelationWeightedProximity cwp( image, imageNew, area2, sigma );
      mm_svd_correspondence< map< int, shared_ptr< IntVector > >::
        const_iterator, CorrelationWeightedProximity >
        corr( cwp, tracking->begin(), tracking->size() );
      corr.match( featuresNew->begin(), featuresNew->size(),
                  insert_iterator< map< int, shared_ptr< IntVector > > >
                  ( *trackingNew, trackingNew->begin() ),
                  back_insert_iterator< vector< shared_ptr< IntVector > > >
                  ( *lostFeatures ) );

    } else
      // Initially schedule all features for assignment of new indices.
      lostFeatures = featuresNew;

    for ( vector< shared_ptr< IntVector > >::iterator i=lostFeatures->begin();
          i != lostFeatures->end(); i++ )
      trackingNew->insert( pair< int, shared_ptr< IntVector > >
                           ( index++, *i ) );

    window->update( *imageNew, trackingNew );
    
    image = imageNew;
    tracking = trackingNew;

  };
  
  } catch ( exception &e ) {
  cerr << endl << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}
