#ifndef MM_CHAMFER_MATCH_H
#define MM_CHAMFER_MATCH_H

#include <cstdlib>
#include <cmath>
#include <ctime>
#include <iostream>
#include <string>
#include <values.h>
#include "mm_config.h"
#include "mm_exception.h"
#include "mm_image.h"
#include "mm_canny.h"
#ifdef HAVE_XALAN
#include "mm_xml_reference.h"
#endif

namespace mimas {
	
/// angle type
enum angle_type {deg, rad};


/** mm_chamfer_match. 
 Object recognition using chamfer matching
 This class performs object recognition by using chamfer maps.
 @author Bala Amavasai (bala@amavasai.org)
 @date Mon Dec  3 19:28:39 GMT 2001
 $Header: /cvs/mimas/include/mm_chamfer_match.h,v 1.19 2005/05/06 09:37:30 engjw Exp $ */
class mm_chamfer_match : public mm_object
{
    private:
  typedef std::string string;
	void paethRotateRadian(mm_image<int> &imagein, mm_image<int> &imageout, float radian); ///< fast rotation using Paeth two shear method
	void invert(mm_image<int> &inimage); ///< invert an image
	void multiLevelCorrelation(mm_image<int> &edgemap, mm_image<int> &objectedgemap, int levels, int &x, int &y, float &corrval); ///< perform multilevel correlation on chamfer map
	void chamfer(mm_image<int> &inimage); ///< convert an edge map to a Chamfer distance map 
#ifdef HAVE_XALAN
   void storeChamferParams( std::ostream &stream ); ///< store object from stream in XML format
   void extractChamferParams( mm_xml_reference< XalanElement > &docElement ) throw (mm_exception); ///< extract object parameters from stream in XML format
#endif 
   string name;	
	float ang1, ang2, angstep;
	float mincorr;
	int mlevels;
	bool chamferModel, sceneAdded, ROI;
	mm_canny<int> canny;
	mm_image<int> scene, model;
	int roi_lx, roi_ly, roi_ux, roi_uy;
	bool cannyAutoThres;
	float cannyStdDev;
	int cannyLowThres, cannyHighThres;


    public:

	mm_chamfer_match(void); ///< constructor
   string getName(void);
	void setName( const std::string &n );
	void setSearchAngleRange(float angle1, float angle2, float anglestep, angle_type ang); ///< rotate the models through angle1 (default=0.0) to angle2 (default=0.0) in steps anglesteps (default=0.0) when performing the search. This facility is due to the fact that chamfer matching is not inherently rotation invariant. angle_type is either {rad, deg}
	void setMultiResLevels(int levels);///< how many levels to use during multiresolution search. Default=1 
	void setChamferModel(bool val); ///< does the model need to be chamfered? Default=true
	bool getChamferModel(void); ///< does the model need to be chamfered?
	void addScene(mm_image<int> &sceneimage); ///< adds/updates scene image.  Note that if the Canny parameters are changed then addScene needs to be called again for the Canny parameters to take effect.
   void findModel(mm_image<int> &modelimage, int &est_x, int &est_y, float &est_anglerad) throw (mm_exception); ///< find the object model given by modelimage in the scene and return the x,y locations and matched angle
	void useROI(bool val); ///< use the ROI when doing recognition
	void setROI(int x1, int y1, int x2, int y2); ///< set the ROI. Activate useROI(true) in order for ROI to be used.	
	void setCannyAutoThres(bool val); ///< autothresholding in Canny stage? Default=true;
	void setCannyStdDev(float val); ///< Canny standard deviation. Default=1.0
	void setCannyLowThres(int val); ///< lower threshold for hysterisis if autothresholding is false
	void setCannyHighThres(int val); ///< upper threshold for hysterisis if autothresholding is false
	float getMinCorrVal(void); ///< get correlation error from last recognition
#ifdef HAVE_XALAN
   void saveChamferParameters( const char *fileName ) throw (mm_exception); ///< save mm_chamfer_match parameters to a file, including model parameters
   void loadChamferParameters( const char *fileName ) throw (mm_exception); ///< load mm_chamfer_match parameters to a file, including model parameters 
#endif
	int track_roi_lx, track_roi_ly, track_roi_ux, track_roi_uy; ///< parameters used for tracking - these does not affect the chamfer matching
	
	bool getCannyAutoThres(void); ///< autothresholding?
	float getCannyStdDev(void); ///< Canny standard deviation. 
	int getCannyLowThres(void); ///< lower threshold for hysterisis if autothresholding is false
	int getCannyHighThres(void); ///< upper threshold for hysterisis if autothresholding is false
	int getMultiResLevels(void);///< how many levels to use during multiresolution search.
	void getSearchAngleRange(float &angle1, float &angle2, float &anglestep); ///< return angles are in radians
	bool getUseROI(void); ///< use the ROI when doing recognition
};

};

#endif
