#include "mm_image.h"
#include "mm_rgba.h"
#include <string>

#ifndef MM_COLOUR_PROPERTIES_TOOL_H
#define MM_COLOUR_PROPERTIES_TOOL_H

namespace mimas {
/**
 * Interpolation/extrapolation method for controlling image properties
 * for more details see: http://www.sgi.com/grafica/interp/index.html
 * @author Bala Amavasai (bala@amavasai.org)
 * @date Tue Aug 20 16:55:28 BST 2002
 * $Header: /cvs/mimas/include/mm_colour_properties_tool.h,v 1.13 2005/04/27 16:19:50 engjw Exp $
 */

template <typename T>
class mm_colour_properties_tool : public mm_object
{
	private:
		void interpolate(mm_image< mm_rgba< T > > &orig, mm_image< mm_rgba< T > > &ref, double weight);
			
	public:
		mm_colour_properties_tool(void) {}
		void brightness(mm_image< mm_rgba< T > > &im, double brightval); ///< brightval is from -1.0 (darkest) to 1.0 (lightest). Values outside this range are extrapolated
		void saturation(mm_image< mm_rgba< T > > &im, double satval); ///< satval is from -1.0 (grayest) to 0.0 (normal) to >0.0 (saturated). Values outside this range are extrapolated
		void contrast(mm_image< mm_rgba< T > > &im, double contval); ///< contval is from -1.0 (lowest) to 0.0 (normal) to >0.0 (contrasty). Values outside this range are extrapolated
		void setChannel(mm_image< mm_rgba< T > > &im, T val, const char *channel); ///< set channel "R","G","B","A" or multiplicates (e.g. "RGB") to val

};

template <typename T>
void mm_colour_properties_tool<T>::interpolate(mm_image< mm_rgba< T > > &orig, mm_image< mm_rgba< T > > &ref, double weight)
{
	double outr, outg, outb;

	for (int j=0; j<orig.getHeight(); j++)
		for (int i=0; i<orig.getWidth(); i++)
		{
         mm_rgba< T > *origPixel = &orig.pixel( i, j );
         const mm_rgba< T > *refPixel = &ref.pixel( i, j );
			outr=(1.0-weight)*(double)refPixel->getRed() + weight*(double)origPixel->getRed();
			outg=(1.0-weight)*(double)refPixel->getGreen() + weight*(double)origPixel->getGreen();
			outb=(1.0-weight)*(double)refPixel->getBlue() + weight*(double)origPixel->getBlue();

			outr=(outr>(T)255)?(T)255:outr;
			outg=(outg>(T)255)?(T)255:outg;
			outb=(outb>(T)255)?(T)255:outb;
			outr=(outr<(T)0)?(T)0:outr;
			outg=(outg<(T)0)?(T)0:outg;
			outb=(outb<(T)0)?(T)0:outb;
			
			origPixel->setRed((T)outr);
			origPixel->setGreen((T)outg);
			origPixel->setBlue((T)outb);
		}
}


template <typename T>
void mm_colour_properties_tool<T>::brightness(mm_image< mm_rgba< T > > &im, double weight)
{
	mm_image< mm_rgba< T > > ref;
	double w;

	//weight=(weight>1.0)?1.0:weight;
	//weight=(weight<-1.0)?-1.0:weight;

   mm_rgba< T > initVal;

	if (weight>=0.0)
	{
		w=1.0-weight;
      initVal = mm_rgba< T >( 255, 255, 255 );
	}
	else
	{
		w=1.0+weight;
      initVal = mm_rgba< T >(   0,   0,   0 );
	}

	ref.init(im.getWidth(),im.getHeight());
   ref.fill( initVal );
	interpolate(im,ref,w);
}

template <typename T>
void mm_colour_properties_tool<T>::saturation(mm_image< mm_rgba< T > > &im, double weight)
{
	mm_image< mm_rgba< T > > ref;
	double w;

	w=weight+1.0;
	
	ref=im;

	double gray;
	for (int j=0; j<ref.getHeight(); j++)
		for (int i=0; i<ref.getWidth(); i++)
		{
        const mm_rgba< T > *pixel  = &ref.pixel( i, j );
        gray = (double)( pixel->getRed() +
                         pixel->getGreen() +
                         pixel->getBlue() )/3.0;
        ref.pixel( i, j ) = mm_rgba< T >( gray, gray, gray );
		}
	
	interpolate(im,ref,w);

}

template <typename T>
void mm_colour_properties_tool<T>::contrast(mm_image< mm_rgba< T > > &im, double weight)
{
	mm_image< mm_rgba< T > > ref;
	double w;

	w=weight+1.0;

	ref.init(im.getWidth(),im.getHeight());
	ref.fill( mm_rgba< T >( 127, 127, 127 ) );

	interpolate(im,ref,w);

}

template <typename T>
void mm_colour_properties_tool<T>::setChannel(mm_image< mm_rgba< T > > &im, T val, const char *channel)
{
	bool R,G,B,A;
   typedef std::string string;
   string c(channel);
	
	R=c.find("R")!=string::npos?true:false;
	G=c.find("G")!=string::npos?true:false;
	B=c.find("B")!=string::npos?true:false;
	A=c.find("A")!=string::npos?true:false;

   
   MMERROR( (int)(R+G+B+A) == (signed)c.length(), mm_exception, ,
            "mm_colour_properties_tool<T>::zeroChannel() - illegal channel "
            "string" );

	for (int j=0; j<im.getHeight(); j++)
		for (int i=0; i<im.getWidth(); i++)
		{
        mm_rgba< T > *pixel = &im.pixel( i, j );
        if (R) pixel->setRed( (T)val );
        if (G) pixel->setGreen( (T)val );
        if (B) pixel->setBlue( (T)val );
        if (A) pixel->setAlpha( (T)val );
		}
}

}

#endif
