#ifndef MM_CORR_TRACKING_H
#define MM_CORR_TRACKING_H

/* #pragma interface */

#include <functional>
#include <iostream>
#include <list>
#include <cstring>
#include <cmath>
#include <values.h>
#include <vector>
#include "mm_cartesian.h"
#include "mm_corrcoef.h"
#include "mm_image.h"
#include <qdom.h>
#include <qfile.h>
#include <qtextstream.h>

namespace mimas {
	
/** Correlation based tracking of gray-level images and templates
    The correlation-type \c C must be an object of class
    \c binary_function with back insertion sequences as first and second
    parameter and double as return value.
    @ingroup featureTracking
    @author Bala Amavasai (bala@amavasai.org)
    @see mm_corrcoef
    @date Fri Apr  6 11:54:53 BST 2001 
    $Header: /cvs/mimas/include/mm_correlation_tracking_tool.h,v 1.7 2005/01/18 18:18:43 engjw Exp $ */
template< typename T, class C = mm_corrcoef< std::vector< double > > >
class mm_correlation_tracking_tool : public mm_object
{
	protected:
		int numPolyPoints; ///< number of points on polygon 
		int *xpoly, *ypoly; ///< point on the polygon 
		std::list< mm_Cartesian<int> > polyCoords, finalPolyCoords; ///< the polygon coordinates before and after execution of tracking procedure
		mm_image<T> trackingTemplate; ///< the template
		mm_image<T> trackingMask; ///< the rectangular mask used to store shape of template 
		int templateRows, templateCols; ///< rows and columns of the bounding box of the polygon template
		mm_Cartesian<int> offsetForFirstCoord; ///< offset of first coordinate point from top left corner of bounding box of template 

	private:
		void findRectTemplate(mm_Cartesian<int> &topLeft, mm_Cartesian<int> &bottomRight);
		int pointInPolygon(int x, int y);
		void * thread(void * ptr); 
		double lastTrackingCorr;
		int templateArea;
		bool autoUpdateTemplate;
		double autoUpdateCorrVal;
		mm_Cartesian<int> trackingRange, startPoint, trackingDispl;
		
	public:
		void pushPolyCoord(int xcoord, int ycoord); ///< push a point of a polygon which specifies the tracking template
		const std::list< mm_Cartesian<int> > &getPolyCoords(void) const; ///< returns the current list of the polygon template points
		void showPolyCoords(void); ///< display the polygon points
		void clearPolyCoords(void); ///< clears the polygon points (e.g. if you want to specify a new polygon)
		void sortPolyCoords(void); ///< sort polygon points (useful when displaying/debugging)
		void createTemplate(mm_image<T> &trackingImage); ///< create the template from the polygon points that have been pushed
		void writeTemplateToFilePGM(char *fn); ///< write the template to a PGM file (for debugging)
		void writeMaskedTemplateToFilePGM(char *fn); ///< write the template mask to a PGM file (for debugging)
		void showPolygonOnImage(mm_image<T> &someImage); ///< display the polygon onto an image
		void showPolygonOnImage(mm_image<T> &someImage, int greylevel); ///< display the polyton onto an image using gray colour \<greylevel\>.
		double getTrackingCorr(void); ///< return the last tracking correlation value
		int getTemplateArea(void); ///< return the area of a template
#ifdef HAVE_XALAN
		void saveParameters( const char *fileName ); ///< save parameters to an XML file
		void loadParameters( const char *fileName ); ///< load parameters from an XML file
#endif
		mm_Cartesian<int> trackCorrelation(mm_image<T> &trackingImage, mm_Cartesian<int> &startCoord, mm_Cartesian<int> &range, C &c ); ///< perform correlation on trackingImage with specified start coordinates and range
		mm_Cartesian<int> trackCorrelation(mm_image<T> &trackingImage, mm_Cartesian<int> &startCoord, C &c );  ///< perform correlation on trackingImage with specified start coordinates. The tracking range is assumed pre-specified using the setTrackingRange() functions.
		mm_Cartesian<int> trackCorrelation(mm_image<T> &trackingImage, C &c ); ///< uses the startCoord from the initial image sent to the template, or the end tracking point from the last tracked image.
		std::list< mm_Cartesian<int> > trackCorrelationList(mm_image<T> &trackingImage, mm_Cartesian<int> &startCoord, mm_Cartesian<int> &range, C &c ); ///< track and return a list of the points of the final polygon after tracking.
		void setAutoUpdateTemplate(bool val); ///< automatically update the template when correlation is to low? Default=false;
		void setAutoUpdateCorrVal(double val); ///< lowest correlation before template is updated.
		void setTrackingRange(int range); ///< tracking range. The single value specifies the ranges in the x and y direction
		void setTrackingRange(int xrange, int yrange);  ///< overloaded function to set x and y tracking range seperately
		mm_Cartesian<int> getTrackingPoint(void);  ///< returns the co-ordinates of the main tracking point
		mm_Cartesian<int> getTrackingDisplacement(void);  ///< returns the last tracking region displacement
		mm_correlation_tracking_tool(void); ///< constructor
		virtual ~mm_correlation_tracking_tool(void); ///< destructor
		void variableDiagnostics(void); ///< stick debugging stuff here
		
		bool debug;
};

}

#include "mm_correlation_tracking_tool.tcc"

#endif
