#include <fftw3.h>
#include <boost/array.hpp>

namespace mimas {

template< typename Real, std::size_t Dim >
boost::multi_array< std::complex< Real >, Dim > mm_fft< Real, Dim >::operator()
  ( const boost::multi_array< std::complex< Real >, Dim > &field )
{
  assert( sizeof( std::complex< Real > ) == sizeof( fftw_complex ) );

  boost::array< std::size_t, Dim > shape;
  boost::copy_n( field.shape(), Dim, shape.begin() );
  
  boost::multi_array< std::complex< Real >, Dim > retVal( shape );

  assert( sizeof( std::size_t ) == sizeof( int ) );
  fftw_plan plan =
    fftw_plan_dft( Dim, (const int *)field.shape(),
                   (fftw_complex *)field.data(),
                   (fftw_complex *)retVal.data(),
                   -1, FFTW_ESTIMATE | FFTW_PRESERVE_INPUT );

  assert( plan != NULL );
  
  fftw_execute( plan );
  fftw_destroy_plan( plan );

  return retVal;
}

template< typename Real, std::size_t Dim >
boost::multi_array< std::complex< Real >, Dim > mm_invfft< Real, Dim >::operator()
  ( const boost::multi_array< std::complex< Real >, Dim > &field )
{
  assert( sizeof( std::complex< Real > ) == sizeof( fftw_complex ) );

  boost::array< std::size_t, Dim > shape;
  boost::copy_n( field.shape(), Dim, shape.begin() );

  boost::multi_array< std::complex< Real >, Dim > retVal( shape );
    
  assert( sizeof( std::size_t ) == sizeof( int ) );
  fftw_plan plan =
    fftw_plan_dft( Dim, (const int *)field.shape(),
                   (fftw_complex *)field.data(),
                   (fftw_complex *)retVal.data(),
                   +1, FFTW_ESTIMATE | FFTW_PRESERVE_INPUT );

  assert( plan != NULL );

  fftw_execute( plan );
  fftw_destroy_plan( plan );

  return retVal * std::complex< Real >( 1.0 / retVal.num_elements() );
}

template< typename Real, std::size_t Dim >
boost::multi_array< std::complex< Real >, Dim > mm_rfft< Real, Dim >::operator()
  ( const boost::multi_array< Real, Dim > &field )
{
  assert( sizeof( std::complex< Real > ) == sizeof( fftw_complex ) );
  assert( field.shape()[ Dim - 1 ] % 2 == 0 );

  boost::array< std::size_t, Dim > shape;
  boost::copy_n( field.shape(), Dim, shape.begin() );
  shape[ Dim - 1 ] = shape[ Dim - 1 ] / 2 + 1;
  
  boost::multi_array< std::complex< Real >, Dim > retVal( shape );

  assert( sizeof( std::size_t ) == sizeof( int ) );
  fftw_plan plan =
    fftw_plan_dft_r2c( Dim, (const int *)field.shape(),
                       (Real *)field.data(),
                       (fftw_complex *)retVal.data(),
                       FFTW_ESTIMATE | FFTW_PRESERVE_INPUT );
  
  assert( plan != NULL );

  fftw_execute( plan );
  fftw_destroy_plan( plan );

  return retVal;
}

template< typename Real, std::size_t Dim >
boost::multi_array< Real, Dim > mm_invrfft< Real, Dim >::operator()
  ( const boost::multi_array< std::complex< Real >, Dim > &field )
{
  // Doesn't preserve input.
  boost::multi_array< std::complex< Real >, Dim > copy( field );
  assert( sizeof( std::complex< Real > ) == sizeof( fftw_complex ) );

  boost::array< std::size_t, Dim > shape;
  boost::copy_n( field.shape(), Dim, shape.begin() );

  shape[ Dim - 1 ] = ( shape[ Dim - 1 ] - 1 ) * 2;
  boost::multi_array< Real, Dim > retVal( shape );
    
  assert( sizeof( std::size_t ) == sizeof( int ) );
  fftw_plan plan =
    fftw_plan_dft_c2r( Dim, (const int *)retVal.shape(),
                       (fftw_complex *)copy.data(),
                       (Real *)retVal.data(),
                       FFTW_ESTIMATE );// FFTW_PRESERVE_INPUT

  assert( plan != NULL );

  fftw_execute( plan );
  fftw_destroy_plan( plan );

  return retVal * ( 1.0 / retVal.num_elements() );
}

}
