//
//
//

//



#ifndef MM_FILTER_H
#define MM_FILTER_H

#include <vector>
#include "mm_exception.h"
#include "mm_image.h"

namespace mimas {
class mm_property_image;

/** @defgroup gradientFilters Gradient filter and other filters
 * General image filter routines. Used with the canny tool and corner tool and
 * gauss tools. These routines operate on a given image and transform it in some
 * way. mm_filter is mainly a namespace used to group filters together.
 *
 *
 * @author Stuart Meikle (stu@stumeikle.org)
 * @date ?? May 2002
 * $Header: /cvs/mimas/include/mm_filter.h,v 1.17 2005/07/27 17:38:38 engjw Exp $
 @{
 */
template<typename T>
class mm_filter
{
 public:
    //just a namespace used to tie all the filters together
    //filters operate on images

protected:
    void	filterRows(mm_image<T> &input, mm_image<T>&output,
                    std::vector<double> * prof1d); ///< fast filter. can be used for symmetric convolution operations
    void	filterCols(mm_image<T> &input, mm_image<T>&output,
                    std::vector<double> * prof1d); ///< fast filter. can be used for symmetric convolution operations
	        //<< convolution functions go here! >>

public:
    //interface
    virtual void filter( mm_image<T>& input )=0;	///< an attempt to provide a standard interface. filter image in situ.
    virtual void filter( mm_image<T>& input, mm_image<T>& output)=0; ///< an attempt to provide a standard interface. filter and put result in output

    //static filters (ie no local vars, can be called without
    //an instance of this class)
    static void gradX( const mm_image<T>& input, mm_image<T>& output ); ///< calculate X direction gradient
    static void gradY( const mm_image<T>& input, mm_image<T>& output ); ///< calculate Y direction gradient
    static void nonMaximaSuppression( const mm_image<T>& gradx,
                                      const mm_image<T>& grady,
                                      const mm_image<T>& gradsq,
                                      double threshold,
                                      mm_property_image& output )
      throw (mm_exception); ///< filter to remove all pixels which are not locally maximal
};

//----------- basic common functionality. should add convolutions --------------

template<typename T>
void mm_filter<T>::filterRows( mm_image<T> &input, mm_image<T>& output,
                               std::vector<double>* prof1d )
{
    double	norm=0;
    int		mp, w=input.getWidth();
    boost::shared_array< T > row( new T[input.getWidth()] );
    double	v;
    int		nx;

    if ( &output != &input )
      output.init( input.getWidth(), input.getHeight());

    for(int i=0;i<(signed)prof1d->size();++i)
	norm+=(*prof1d)[i];

    if (norm==0)	return;
    mp = prof1d->size()/2;

    for(int y=0;y<input.getHeight();++y)
    {
        //stumeikle20030304 break the operation into 3 parts now...
	//first bit...
	for(int x=0;x<mp;++x)
	{
	    v=0;

	    nx=x-mp;
	    for(int i=0;i<(signed)prof1d->size();++i,++nx)
	    {
		if (nx>0 && nx<w)
		{
		    v+=input.pixel(nx,y) * (*prof1d)[i];
		}
	    }
	    v/=norm;
	    row[x] = (T)v;
	}

	//main bit..
	for(int x=mp;x<(w-mp);++x)
	{
	    v=0;

	    nx=x-mp;
	    for(int i=0;i<(signed)prof1d->size();++i,++nx)
		v+=input.pixel(nx,y) * (*prof1d)[i];
	    v/=norm;
	    row[x] = (T)v;
	}

	//end bit
	for(int x=w-mp;x<w;++x)
	{
	    v=0;

	    nx=x-mp;
	    for(int i=0;i<(signed)prof1d->size();++i,++nx)
	    {
		if (nx>0 && nx<w)
		{
		    v+=input.pixel(nx,y) * (*prof1d)[i];
		}
	    }
	    v/=norm;
	    row[x] = (T)v;
	}

	//copy the row to the output
   copy_n( row.get(), input.getWidth(), output.rawData()[y].begin() );
    }
}


template<typename T>
void mm_filter<T>::filterCols( mm_image<T> &input, mm_image<T>& output,
                               std::vector<double>* prof1d )
{
    double	norm=0;
    int		i,mp, h=input.getHeight(),y;
    boost::shared_array< T > col( new T[input.getHeight()] );
    double 	v=0;
    int	   	ny;


    if ( &output != &input )
      output.init( input.getWidth(), input.getHeight());

    for(i=0;i<(signed)prof1d->size();++i)
	norm+=(*prof1d)[i];

    if (norm==0)	return;
    mp = prof1d->size()/2;

    for(int x=0;x<input.getWidth();++x)
    {
	for(y=0;y<h;++y)
	    col[y]=input.pixel(x,y);

	//you could use the tina optimations here but they're too :-*
	//break this into 3 parts now. stumeikle 20030304
	//the start...
	for(y=0;y<mp;++y)
	{
	    v=0;

	    ny=y-mp;
	    for(i=0;i<(signed)prof1d->size();++i,++ny)
	    {
		if (ny>0 && ny<h)
		    v+=input.pixel(x,ny) * (*prof1d)[i];
	    }
	    v/=norm;
	    col[y] = (T)v;
	}

	//middle part
	for(y=mp;y<(h-mp);++y)
	{
	    v=0;

	    ny=y-mp;
	    for(i=0;i<(signed)prof1d->size();++i,++ny)
		    v+=input.pixel(x,ny) * (*prof1d)[i];
	    v/=norm;
	    col[y] = (T)v;
	}

	//end bit
	for(y=(h-mp);y<h;++y)
	{
	    v=0;

	    ny=y-mp;
	    for(i=0;i<(signed)prof1d->size();++i,++ny)
	    {
		if (ny>0 && ny<h)
		    v+=input.pixel(x,ny) * (*prof1d)[i];
	    }
	    v/=norm;
	    col[y] = (T)v;
	}

	for(y=0;y<h;++y)
	    output.pixel(x,y)=col[y];
    }

}
///@}

}

#include "mm_filter_grad.h"
#include "mm_filter_nonmaxsup.h"

#endif
