#ifndef MM_GAUSS_H
#define MM_GAUSS_H

#include <boost/multi_array.hpp>
#include <cassert>
#include <cmath>
#include <deque>
#include "mm_image.h"
#include "mm_image_conv.h"

namespace mimas {

/** @defgroup gauss Gaussian blur
    Methods for blurring images with a gauss-bell. The amount of blurring can
    be specified and the size of the filter is computed by choosing an upper
    bound for the approximation-error.
    @author Stuart Meikle (stu@stumeikle.org)
    @author Jan Wedekind (jan@wedesoft.de)    
    @date Fri Jul 22 14:08:14 2005
    @{ */
/** Compute gauss-bell.
    The values of the cells are computed by using the integral of the
    gauss-function:
    \f$\displaystyle\int_{-r}^{+r}{\frac{1}{\sqrt{2\,\pi}\,\sigma}\,e^{-\displaystyle\frac{x^2}{2\,\sigma^2}}\,\mathrm{d}x}=\mathrm{erf}(\displaystyle\frac{r}{\sqrt{2}\,\sigma})\f$
    @param sigma Standard deviation of gauss-distribution.
    @param maxError Maximum error boundary
    (relative to range of pixel-values).
    @see gaussian_blur */
template< typename T >
std::deque< T > gauss_bell( T sigma, T maxError = (T)( 1.0 / 256.0 ) );

/** Blur 2-D array.
    Perform gaussian blur on 2-D array.
    @param x Input array.
    @param sigma Standard deviation.
    @param maxError Maximum error boundary (relative to range of pixel-values).
    @see gauss_bell */
template< typename T >
boost::multi_array< T, 2 > gaussian_blur
  ( const boost::const_multi_array_ref< T, 2 > &x, T sigma, T maxError );

/** Blur image.
    Perform gaussian blur on 2-D array.
    @param x Input image.
    @param sigma Standard deviation.
    @param maxError Maximum error boundary (relative to range of pixel-values).
    @see gauss_bell */
template< typename T >
mm_image< T > gaussian_blur( const mm_image< T > &x, T sigma,
                            T maxError = (T)( 1.0 / 256.0 ) )
{
  return mm_image< T >( gaussian_blur< T >( x.rawData(), sigma, maxError ) );
}

///@}

};

#include "mm_gauss.tcc"

#endif
