namespace mimas {

template< typename T >
std::deque< T > gauss_bell( T sigma, T maxError )
{
  assert( sigma > 0 );

  std::deque< T > retval;

  // Integral of gauss-curve from -0.5 to +0.5.
  T integral = erf( 0.5 / ( sqrt( 2.0 ) * sigma ) );
  retval.push_back( integral );

  while ( 1.0 - integral > maxError ) {
    // Compute integral of gauss-curve from -newSize2 to +newSize2.
    const T
      newSize2= 0.5 * ( retval.size() + 2 ),
      newIntegral = erf( newSize2 / ( sqrt( 2.0 ) * sigma ) ),
      value = 0.5 * ( newIntegral - integral );
    retval.push_back( value );
    retval.push_front( value );
    integral = newIntegral;
  };

  // Normalise to maintain power of image.
  const T factor = 1.0 / integral;
  for ( typename std::deque< T >::iterator i = retval.begin();
        i != retval.end(); i++ )
    *i *= factor;

  return retval;
}

template< typename T >
boost::multi_array< T, 2 > gaussian_blur
  ( const boost::const_multi_array_ref< T, 2 > &x, T sigma, T maxError )
{
  // Split error-budget up for two operations.
  std::deque< T > filter( gauss_bell< T >( sigma, maxError / 2.0 ) );
  boost::multi_array< T, 2 >
    filterx( boost::extents[ 1            ][ filter.size() ] ),
    filtery( boost::extents[ filter.size()][ 1             ] );
  std::copy( filter.begin(), filter.end(), filterx.data() );
  std::copy( filter.begin(), filter.end(), filtery.data() );
  return correlate( correlate( x, filterx ), filtery );
}
   
};
