#ifndef MM_IMAGE_FILEOUTPUT_H
#define MM_IMAGE_FILEOUTPUT_H

#include <iostream>
#include "mm_image_output.h"
#include "mm_rgba.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @addtogroup imageOutput
    @{ */
/** Sink for writing an image to a file.
    The stream is used for writing image(s) to. The stream-manipulator
    \c setformat can be used to change the file-format.
    Note, that \c mm_image_fileinput only allows reading a single image from
    a stream.
    @see mm_image_fileoutput 
    @see setformat */
template< typename T >
class mm_image_fileoutput: public mm_image_output< T >
{
 public:
  /** Constructor.
      @param _stream Stream to write images to. */
  explicit mm_image_fileoutput( std::ostream &_stream ): stream( &_stream ) {}
  /** Write image to stream.
      @param image Image to write to the stream. */
  virtual void write( const mm_image< T > &image );
 protected:
  /// Pointer to stream.
  std::ostream *stream;
};

template<>
void mm_image_fileoutput< unsigned char >::write
  ( const mm_image< unsigned char > &image );
template<>
void mm_image_fileoutput< mm_rgba< unsigned char > >::write
  ( const mm_image< mm_rgba< unsigned char > > &image );

/// File-format of next image to be written.
extern const char *mm_image_fileoutput_fileType;

///
struct _format { const char *name; };
/** Stream-manipulator to change the file-format.
    The Qt-documentation contains information about what file formats are
    supported (class \c QImageIO). After the image was written, the default
    value will be restored.
    @see mm_image_fileoutput
    @see mm_image_fileoutput_fileType */
inline _format setformat( const char *name )
{ _format retVal; retVal.name = name; return retVal; }

/** Change the file-format for the next image.
    Unfortunately this method is not side-effect free.
    @see mm_image_fileoutput_fileType */
inline std::ostream &operator<<( std::ostream &stream, _format format )
{ mm_image_fileoutput_fileType = format.name; return stream; }

/// Stream-operator for writing image to a stream.
template< typename T, typename U >
inline mm_image_fileoutput< T > &operator<<( mm_image_fileoutput< T > &output,
                                             const mm_image< U > &image )
{
  if ( typeid( T ) == typeid( U ) )
    output.write( (mm_image< T > &)image );
  else {
    output.write( mm_image< T >( image ) );
  };
  return output;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< unsigned char > &img )
{
  mm_image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< int > &img )
{
  mm_image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< double > &img )
{
  mm_image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< float > &img )
{
  mm_image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< mm_rgba< unsigned char > > &img )
{
  mm_image_fileoutput< mm_rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< mm_rgba< int > > &img )
{
  mm_image_fileoutput< mm_rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< mm_rgba< float > > &img )
{
  mm_image_fileoutput< mm_rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const mm_image< mm_rgba< double > > &img )
{
  mm_image_fileoutput< mm_rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

///@}

///@}

};

#endif
