namespace mimas {

template< typename T >
mm_image_firewireinput< T >::mm_image_firewireinput( int _mode,
                                                     int _framerate,
                                                     int _node,
                                                     const char *_device,
                                                     int _channel,
                                                     int _speed )
  throw (mm_exception):
  mode(_mode), node(_node)
{
  handle = dc1394_create_handle( 0 );
  MMERROR( handle != NULL, mm_exception, ,
           "Unable to acquire raw1394 handle. Please check, wether the kernel "
           "modules `ieee1394',`raw1394' and `ohci1394' are loaded and wether "
           "you have read/write permission on /dev/raw1394." );
  try {

    int numNodes = raw1394_get_nodecount(handle);
    int numCameras;
#ifndef NDEBUG
    nodeid_t *cameraNode = dc1394_get_camera_nodes( handle, &numCameras, 1 );
#else
    nodeid_t *cameraNode = dc1394_get_camera_nodes( handle, &numCameras, 0 );
#endif
    MMERROR( numCameras >= 0, mm_exception, ,
             "Failed to acquire list of camera nodes." );
    MMERROR( numCameras >= 1, mm_exception, ,
             "Did not find a single digital camera on the firewire bus." );

    // Check for iso-transfer bug of old raw1394 system
    // (before kernel 2.4.21-pre2).
    // Compare http://sourceforge.net/tracker/index.php?func=detail&aid=435107&group_id=8157&atid=108157
    for ( int i=0; i<numCameras; i++ ) {
      MMERROR( cameraNode[i] != numNodes - 1, mm_exception, ,
               "A camera has become the root-node of the firewire bus. Try "
               "to fix this problem by reloading the `ohci1394'-module with "
               "option `attempt_root=1'." );
    };

    MMERROR( node >= 0 && node < numCameras, mm_exception, ,
             "Camera-node number " << node << " out of range. The range is "
             "[ 0; " << numCameras << " )." );

    // Get list of possible framerates.
    quadlet_t frameRates;
    MMERROR( dc1394_query_supported_framerates( handle, cameraNode[ node ],
                                                FORMAT_VGA_NONCOMPRESSED,
                                                mode,
                                                &frameRates )
             == DC1394_SUCCESS, mm_exception, ,
             "Requested combination of mode and format doesn't seem to be "
             "supported." );
    
    // Check, wether requested framerate is supported.
    MMERROR( ( frameRates & ( 1U << ( 31 + FRAMERATE_MIN - _framerate ) ) )
             != 0, mm_exception, ,
             "Specified framerate not supported by camera." );

    // Setup DMA capture (allocates ringbuffer).
    MMERROR( dc1394_dma_setup_capture( handle,
                                       cameraNode[ node ],
                                       _channel,
                                       FORMAT_VGA_NONCOMPRESSED,
                                       mode,
                                       _speed,
                                       _framerate,
                                       4,// Number of dma-buffers.
#ifdef LIBDC1394_OLD
                                       0,// Extra-buffering (?)
#endif
                                       1,// Drop frames.
                                       _device,// Device (/dev/video1394)
                                       &camera ) == DC1394_SUCCESS,
             mm_exception, ,
             "Error setting up DMA capture." );

    try {

      // Start capture.
      MMERROR( dc1394_start_iso_transmission( handle, camera.node )
               == DC1394_SUCCESS, mm_exception, ,
               "Error starting camera iso transmission." );
      
    } catch ( mm_exception &e ) {

      // Free ringbuffer.
      dc1394_dma_release_camera( handle, &camera );
      throw e;

    };

  } catch ( mm_exception &e ) {

    // Free handle.
    dc1394_destroy_handle( handle );
    throw e;

  };
}

template< typename T >
mm_image_firewireinput< T >::~mm_image_firewireinput(void)
{
  // Halt iso reception.
  dc1394_dma_unlisten( handle, &camera );
  // Stop iso transmission.
  dc1394_stop_iso_transmission( handle, camera.node );
  // Free mapped memory.
  dc1394_dma_release_camera( handle, &camera );
  // Free handle.
  dc1394_destroy_handle( handle );
}

template< typename T >
dc1394_feature_info mm_image_firewireinput< T >::get_feature( int _id )
  throw (mm_exception)
{
  dc1394_feature_info retval; retval.feature_id = _id;
  MMERROR( dc1394_get_camera_feature( handle, node, &retval ) != 0,
           mm_exception, ,
           "Error requesting feature " << _id << " from firewire "
           "digital camera." );
  return retval;
}

template< typename T >
unsigned int mm_image_firewireinput< T >::get_feature_value( int _id )
  throw (mm_exception)
{
  unsigned int retval;
  MMERROR( dc1394_get_feature_value( handle, node, _id, &retval ) != 0,
           mm_exception, ,
           "Error requesting value of feature " << _id << " from firewire "
           "digital camera." );
  return retval;
}

template< typename T >
 void mm_image_firewireinput< T >::set_feature_value( int _id,
                                                      unsigned int _value )
  throw (mm_exception)
{
  MMERROR( dc1394_set_feature_value( handle, node, _id, _value ) != 0,
           mm_exception, ,
           "Error setting feature " << _id << " from firewire "
           "digital camera." );
}

};
