//
// Generic mm_image functions
//
// Bala Amavasai (bala@amavasai.org)
// Mon Oct  1 15:41:44 BST 2001
//
// $Header: /cvs/mimas/include/mm_image_funcs.h,v 1.39 2005/08/18 17:19:38 engmb Exp $
//

#ifndef MM_IMAGE_FUNCS_H
#define MM_IMAGE_FUNCS_H

#include <cassert>
#include <iostream>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <values.h>
#include <cmath>
#include "mm_image.h"
#include "mm_image_conv.h"
#include "mm_image_op.h"
#include "mm_image_qtoutput.h"
#include "mm_rgba.h"
#include "mm_prop.h"
#include "mm_property_image.h"
#include "mm_image_draw.h"

namespace mimas {

template< typename T >
T min_val( const mm_image< T > &image )
{
  T retVal;
  if ( image.initialised() )
    retVal = *min_element( image.rawData().data(),
                           image.rawData().data() +
                           image.rawData().num_elements() );
  else
    retVal = T();
  return retVal;
}

template< typename T >
T max_val( const mm_image< T > &image )
{
  T retVal;
  if ( image.initialised() )
    retVal = *max_element( image.rawData().data(),
                           image.rawData().data() +
                           image.rawData().num_elements() );
  else
    retVal = T();
  return retVal;
}

template< typename T >
mm_image< T > &normaliseIt( mm_image< T > &image,
                            const T &val1, const T &val2 )
{
  if ( image.initialised() ) {// Image must not be empty.
    T
      minval = min_val( image ),
      maxval = max_val( image );
    _normalise< T > n( minval, maxval, val1, val2 );
    mm_array_apply_intern< _normalise< T >, 2 > apply( n );
    apply( image.rawData().origin(),
           image.rawData().shape(),
           image.rawData().strides() );
  };
  return image;
}

template< typename T >
mm_image< T > normalise( const mm_image< T > &image,
                         const T &val1, const T &val2 )
{
  mm_image< T > retVal;
  if ( image.initialised() ) {// Image must not be empty.
    retVal.init( image.getWidth(), image.getHeight() );
    T
      minval = min_val( image ),
      maxval = max_val( image );
    _normalise< T > n( minval, maxval, val1, val2 );
    mm_array_apply_extern< _normalise< T >, 2 > apply( n );
    apply( retVal.rawData().origin(), image.rawData().origin(),
           retVal.rawData().shape(), image.rawData().shape(),
           retVal.rawData().strides(), image.rawData().strides() );
  };
  return retVal;
}

template<typename T>
void equalise(mm_image<T> &imagein)
{

    int H[256], TR[256];

    for (int count=0; count<256; count++)
	H[count]=0;

    int intensity;
    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
	{
	    intensity=imagein.pixel(i,j);
       assert( intensity <= 255 );
	    H[intensity]+=1;
	}
    
   // convert to cumulative histogram 
    int imagesize=imagein.getWidth()*imagein.getHeight();
    for (int count=1; count<256; count++)
    {
	H[count]=H[count]+H[count-1];
	TR[count-1]=(255*H[count-1])/imagesize;
    }
    TR[255]=(255*H[255])/imagesize;
     
    for (int j=0; j<imagein.getHeight(); j++)
        for (int i=0; i<imagein.getWidth(); i++)
          imagein.pixel(i,j)=(T)TR[imagein.pixel(i,j)];

}

template< typename T >
mm_image< T > &bilevelIt( mm_image< T > &image, T threshval,
                          T val1, T val2 )
{
  _bilevel< T > b( val1, val2 );
  mm_array_apply_intern< std::binder2nd< _bilevel< T > >, 2 >
    apply( std::bind2nd( b, threshval ) );
  apply( image.rawData().origin(),
         image.rawData().shape(),
         image.rawData().strides() );
  return image;
}

template< typename T >
mm_image< T > bilevel( const mm_image< T > &image, T threshval,
                       T val1, T val2 )
{
  mm_image< T > retVal; retVal.init( image.getWidth(), image.getHeight() );
  _bilevel< T > b( val1, val2 );
  mm_array_apply_extern< std::binder2nd< _bilevel< T > >, 2 >
    apply( std::bind2nd( b, threshval ) );
  apply( retVal.rawData().origin(), image.rawData().origin(),
         retVal.rawData().shape(), image.rawData().shape(),
         retVal.rawData().strides(), image.rawData().strides() );
  return retVal;
}

template<typename T, typename U>
void despeckleKuwahara(const mm_image<T> &imagein, mm_image<U> &imageout)
{
  if ( &imageout != &imagein )
    imageout.init(imagein.getWidth(),imagein.getHeight());
    
    U intensity;
    float m_ul, m_ur, m_ll, m_lr;
    float v_ul, v_ur, v_ll, v_lr;
    
    for (int y=0; y<imagein.getHeight(); y++)
	for (int x=0; x<imagein.getWidth(); x++)
	{
	    if (y<2 || y>imagein.getHeight()-3 || x<2 || x>imagein.getWidth()-3)
	    {
         imageout.pixel(x,y)=imagein.pixel(x,y);
		continue;

	    }


	    // mean
	    m_ul=0.0; m_ur=0.0; m_ll=0.0; m_lr=0.0;

	    for (int j=0; j<3; j++)
		for (int i=0; i<3; i++)
		{
		    m_ul+=(float)imagein.pixel(x-2+i,y-2+j);
		    m_ur+=(float)imagein.pixel(x+i,y-2+j);
		    m_ll+=(float)imagein.pixel(x-2+i,y+j);
		    m_lr+=(float)imagein.pixel(x+i,y+j);
		}

	    m_ul=m_ul/9.0;
	    m_ur=m_ur/9.0;
	    m_ll=m_ll/9.0;
	    m_lr=m_lr/9.0;

	    
	    // estimated variance
	    v_ul=0.0; v_ur=0.0; v_ll=0.0; v_lr=0.0;
	    
	    for (int j=0; j<3; j++)
		for (int i=0; i<3; i++)
		{
		    v_ul+=(((float)imagein.pixel(x-2+i,y-2+j)-m_ul)*((float)imagein.pixel(x-2+i,y-2+j)-m_ul));
		    v_ur+=(((float)imagein.pixel(x+i,y-2+j)-m_ur)*((float)imagein.pixel(x+i,y-2+j)-m_ur));
		    v_ll+=(((float)imagein.pixel(x-2+i,y+j)-m_ll)*((float)imagein.pixel(x-2+i,y+j)-m_ll));
		    v_lr+=(((float)imagein.pixel(x+i,y+j)-m_lr)*((float)imagein.pixel(x+i,y+j)-m_lr));
		}

	    intensity=(U)m_ul;
	    intensity=v_ur<v_ul?(U)m_ur:intensity;
	    intensity=v_ll<v_ur?(U)m_ll:intensity;
	    intensity=v_lr<v_ll?(U)m_lr:intensity;

	    imageout.pixel(x,y)=intensity;
	}  
}



/**
 *edge detector, can be used to as an example how to use the iterator 
 *
 *Do a query replace
 *Here are the meaning of the abreviations:  
 *ul: uperLeft
 *lr: lowerRight
 *w: width
 *h: height
 *is: iterator_source
 *id: iterator_destination
 *ixs: line_iterator_source
 *ixd: line_iterator_destination
 *t: threshold
 *
 *@author Manuel Boissenin
 *@date 09/03/2005
 * Copyright (C) 2005 by Manuel Boissenin
 ***********************************************************************/

template<typename ImageIteratorSource>
mm_image<typename ImageIteratorSource::value_type> gradient(ImageIteratorSource ul, ImageIteratorSource lr, unsigned int t)
{  
  typedef typename ImageIteratorSource::value_type pixel_type;
  unsigned int w = lr.x - ul.x, h = lr.y - ul.y;

  mm_image<pixel_type> gradient_image;

  gradient_image.init(w, h);
  
  ImageIteratorSource is(ul);
  ImageIteratorSource id = gradient_image.ul();

  //!!for the moment we don't take into account the border cases
  is.y++,id.y++; 
  for(unsigned int y = 1; y< h - 1; y++, is.y++, id.y++)
    {
      //create an iterator for each line
      ImageIteratorSource ixs(is);
      ImageIteratorSource ixd(id);

      //!!for the moment we don't take into account the border cases
      ixs.x++,ixd.x++;

      for (unsigned int x = 1; x < w - 1; x++, ixs.x++, ixd.x++)
	{		
	  const ImageIteratorSource ref(ixs);

	  int temp1 = ref[Dist2D(0,-1)] - ref[Dist2D(0,1)];
	  int temp2 = ref[Dist2D(-1,0)] - ref[Dist2D(1,0)];
	  	  
	  if(temp1 < 0)
	    temp1 = -temp1;
	   if(temp2 < 0)
	    temp2 = -temp2;

	   if(unsigned(temp1 + temp2) > t)
	     {
	       *(ixd) = pixel_type(255);
	     }
	}      
    }  
  return gradient_image;
}

/**
 *Generally the edges are the features considered to caracterize an object.
 *However in certain situations, uniform patchs are caracteristic
 *of the object. In this case a uniform patch is a rectangular region where 
 *all pixel differ little from the central pixel. 
 *The notion can of course be extended to other shapes (a way to do this 
 *would be to use weights) and so on.
 *
 *The size of the patch is 2 * patch_mid_size + 1
 *
 *
 *The algorithm works for color image as well as for grey scale image however
 *I don't understand and haven't check the mechanism that allows the function 
 *to do operations such as additions on mm_rgba<unsigned char> class for instance.
 *
 *@author Manuel Boissenin
 *@date 09/03/2005
 *Copyright (C) 2005 by Manuel Boissenin
 ***********************************************************************/



template<typename ImageIteratorSource>
mm_image<typename ImageIteratorSource::value_type> uniformPatch(ImageIteratorSource ul, ImageIteratorSource lr, unsigned int patch_mid_size, float t)
{  
  typedef typename ImageIteratorSource::value_type pixel_type;
  unsigned int w = lr.x - ul.x, h = lr.y - ul.y;

  mm_image<pixel_type> patch_image;
  patch_image.init(w, h);
  
  ImageIteratorSource is(ul);
  ImageIteratorSource id = patch_image.ul();

  //!!for the moment we don't take into account the border cases
  is.y++,id.y++; 
  for(unsigned int y = patch_mid_size; y< h - patch_mid_size; y++, is.y++, id.y++)
    {
      //create an iterator for each line
      ImageIteratorSource ixs(is);
      ImageIteratorSource ixd(id);

      //!!for the moment we don't take into account the border cases
      ixs.x++,ixd.x++;

      for (unsigned int x = patch_mid_size; x < w - patch_mid_size; x++, ixs.x++, ixd.x++)
	{		
	  const ImageIteratorSource ref(ixs);


	  float sum;
	  for (int x = -patch_mid_size; x <= signed(patch_mid_size); x++)
	    for (int y = -patch_mid_size; y <= signed(patch_mid_size); y++)
	      {
		int temp = ref[Dist2D(x,y)] - ref[Dist2D(0,0)];
		if(temp<0){temp=-temp;}
		sum += temp;
	      }

	  //so we can think of the difference as it it was concerning one pixel
	  sum /= float((patch_mid_size * 2) + 1);   

	  
	  if(sum < t)
	     {
	       *(ixd) = pixel_type(255);
	     }
	}      
    }  
  return patch_image;
}



template<typename T>
void despeckleKuwahara(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    despeckleKuwahara(imagein, tempimage);

    imagein = tempimage;
}

template< typename T >
mm_image< T > gradSobelX( const mm_image< T > &image )
{
  const T sobelx1[1][3]={ {-1, 0, 1} };
  const T sobelx2[3][1]={ { 1 },
                          { 2 },
                          { 1 } };
  boost::const_multi_array_ref< T, 2 > filter1( &sobelx1[0][0],
                                                boost::extents[1][3] );
  boost::const_multi_array_ref< T, 2 > filter2( &sobelx2[0][0],
                                                boost::extents[3][1] );
  return mm_image< T >
    ( correlate( correlate( image.rawData(), filter1 ), filter2 ) );
}


template< typename T >
mm_image< T > gradSobelY( const mm_image< T > &image )
{
  const T sobely1[3][1]={ { -1 },
                          {  0 },
                          {  1 } };
  const T sobely2[1][3]={ { 1, 2, 1} };
  boost::const_multi_array_ref< T, 2 > filter1( &sobely1[0][0],
                                                boost::extents[3][1] );
  boost::const_multi_array_ref< T, 2 > filter2( &sobely2[0][0],
                                                boost::extents[1][3] );
  return mm_image< T >
    ( correlate( correlate( image.rawData(), filter1 ), filter2 ) );
}

template< typename T >
mm_image< T > edgeSobel2( const mm_image< T > &image )
{
  mm_image< T >
    gradX( gradSobelX( image ) ),
    gradY( gradSobelY( image ) );
  return gradX * gradX + gradY * gradY;
}

template< typename T >
mm_image< T > edgeSobel( const mm_image< T > &image )
{
  return squareRoot( edgeSobel2( image ) );
}

// template<typename T, typename U>
// void edgeSobelOrientation(const mm_image<T> &imagein, mm_image<U> &imageout)
// {
//     int	x,y, sx,sy;
//     int	height, width;
//     mm_image<U> tempimage;
//     float	sxIntensity, syIntensity;


//     static float sobelx[3][3]={	{ -1.0, -2.0,-1.0},
//                                 {  0.0,  0.0, 0.0},
//                                 {  1.0,  2.0, 1.0}};

//     static float sobely[3][3]={ 	{-1.0, 0.0, 1.0},
//                                  {-2.0, 0.0, 2.0},
//                                  {-1.0, 0.0, 1.0}};

//     height=imagein.getHeight();
//     width=imagein.getWidth();
//     tempimage.init(width,height);

//     for(y=1;y<height-1;y++)
//         for (x=1;x<width-1;x++)
//         {
//             sxIntensity=0.0; syIntensity=0.0;
//             for(sx=0;sx<3;sx++)
//                 for(sy=0;sy<3;sy++)
//                 {
//                     sxIntensity+=sobelx[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
//                     syIntensity+=sobely[sy][sx]*(float)imagein.pixel(x+sx-1,y+sy-1);
//                 }
//             if (sxIntensity==0.0)
//               tempimage.pixel(x,y)=0.0;
//             else
//               tempimage.pixel(x,y)=atan(syIntensity/sxIntensity);


//         }

//     imageout = tempimage;
// }

template< typename T >
mm_image< T > edgeLaplacian( const mm_image<T> &image )
{
  const T laplacian[3][3]={ {  0, -1,  0 },
                            { -1,  4, -1 },
                            {  0, -1,  0 } };
  boost::const_multi_array_ref< T, 2 > filter( &laplacian[0][0],
                                               boost::extents[3][3] );
  return mm_image< T >( correlate( image.rawData(), filter ) );
}

template< typename T >
mm_image< T > edgeLoG( const mm_image<T> &imagein )
{
  const T LoG[9][9]= { {   0,  0,  3,  2,  2,  2,  3,  0,  0},
                       {   0,  2,  3,  5,  5,  5,  3,  2,  0},
                       {   3,  3,  5,  3,  0,  3,  5,  3,  3},
                       {   2,  5,  3,-12,-23,-12,  3,  5,  2},
                       {   2,  5,  0,-23,-40,-23,  0,  5,  2},
                       {   2,  5,  3,-12,-23,-12,  3,  5,  2},
                       {   3,  3,  5,  3,  0,  3,  5,  3,  3},
                       {   0,  2,  3,  5,  5,  5,  3,  2,  0},
                       {   0,  0,  3,  2,  2,  2,  3,  0,  0} };

  boost::const_multi_array_ref< T, 2 > filter( &LoG[0][0],
                                               boost::extents[9][9] );
  mm_image< T > tempimage( correlate( imagein.rawData(), filter ) );
    
  /*static float LoG[5][5]={{  0.0, 0.0,-1.0, 0.0, 0.0},
                            {  0.0,-1.0,-2.0,-1.0, 0.0},
                            { -1.0,-2.0,16.0,-2.0,-1.0},
                            {  0.0,-1.0,-2.0,-1.0, 0.0},
                            {  0.0, 0.0,-1.0, 0.0, 0.0}};
  */

  bilevelIt( tempimage, (T)0, (T)255, (T)0 );
  
  mm_image< T > imageout;
  imageout.init( imagein.getWidth(), imagein.getHeight() );

  // Shouldn't actually threshold - must check zero crossings
  for ( int y=4; y<imagein.getHeight()-4; y++ )
    for ( int x=4; x<imagein.getWidth()-4; x++ ) {
      if (tempimage.pixel(x,y)<=127) continue;
      if ( ( tempimage.pixel( x-1, y-1 ) > 127 ) ||
           ( tempimage.pixel( x-1, y   ) > 127 ) ||
           ( tempimage.pixel( x-1, y+1 ) > 127 ) ||
           ( tempimage.pixel( x  , y-1 ) > 127 ) ||
           ( tempimage.pixel( x  , y+1 ) > 127 ) ||
           ( tempimage.pixel( x+1, y-1 ) > 127 ) ||
           ( tempimage.pixel( x+1, y   ) > 127 ) ||
           ( tempimage.pixel( x+1, y+1 ) > 127 ) )
        imageout.pixel( x  , y   ) = 255;
    }
  return imageout;
}

// IMPORTANT NOTE
// imagein1 is image at time t-1
// imagein2 is image at time t
template<typename T>
mm_image< T > edgeHaynesJain( const mm_image<T> &imagein1,
                              const mm_image<T> &imagein2 )
{
    MMERROR( imagein2.getWidth () == imagein1.getWidth () &&
             imagein2.getHeight() == imagein1.getHeight(),
             mm_exception, , "mm_image_funcs::edgeHaynesJain - "
             "size of input images do not match" );

    return absolute<T>( edgeSobel<T>( imagein2 ) * ( imagein2 - imagein1 ) );
}


template<typename T, typename U>
void halfResolution(const mm_image<T> &imagein, mm_image<U> &imageout)
{
    // note that this is only an estimate!
    static float gauss[3][3]={	{  1.0,  2.0, 1.0},
                               {  2.0,  4.0, 2.0},
                               {  1.0,  2.0, 1.0}};
    static float gaussfac=(1.0/16.0);

    assert( &imageout != &imagein );
    imageout.init(imagein.getWidth()/2, imagein.getHeight()/2);

    for (int j=1; j<imagein.getHeight(); j+=2)
        for (int i=1; i<imagein.getWidth(); i+=2)
        {
            float tempval=0.0;
            for (int y=-1; y<=1; y++)
                for (int x=-1; x<=1; x++)
                    tempval+=(imagein.pixel(i+x,j+y)*gauss[y+1][x+1]);
            imageout.pixel(i/2,j/2)=(U)(tempval*gaussfac);
        }
}


template<typename T>
void halfResolution(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    halfResolution(imagein, tempimage);

    imagein = tempimage;
}

// Note this assumes black edges on white background unlike halfResolution
// above
template<typename T, typename U>
void halfResolutionEdgePyramid(const mm_image<T> &imagein, mm_image<U> &imageout)
{
  // set background colour of object to be white
  assert( &imageout != &imagein );
  imageout.init(imagein.getWidth()/2, imagein.getHeight()/2);
  imageout.fill(255);

    for (int j=1; j<imagein.getHeight(); j+=2)
        for (int i=1; i<imagein.getWidth(); i+=2)
        {
            int tempval=0;
	    for (int y=-1; y<=1; y++)
                for (int x=-1; x<=1; x++)
                    if (imagein.getPixel(i+x,j+y)>127)
			tempval=255;
	    
	    imageout.pixel(i/2,j/2)=(U)tempval;
        }
}


template<typename T>
void halfResolutionEdgePyramid(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    halfResolutionEdgePyramid(imagein, tempimage);

    imagein = tempimage;
}

// 3x3 median despecking filter
template<typename T>
void despeckleMedian(const mm_image<T> &imagein, mm_image<T> &imageout)
{
  std::vector<T> pixvals;
  assert( &imageout != &imagein );
  imageout.init(imagein.getWidth(),imagein.getHeight());
 
  for (int j=1; j<imagein.getHeight()-1; j++)
	for (int i=1; i<imagein.getWidth()-1; i++)
	{
	    if (!pixvals.empty()) pixvals.clear();
	    for (int y=-1; y<=1; y++)
		for (int x=-1; x<=1; x++)
		    pixvals.push_back(imagein.pixel(i+x,j+y));
	    
	    sort(pixvals.begin(), pixvals.end());
	    imageout.pixel(i,j)=pixvals[pixvals.size()/2];
	}
 
    for (int j=0; j<imagein.getHeight(); j++)
    {
      imageout.pixel(0,j) = imagein.pixel(0,j);
      imageout.pixel(imagein.getWidth()-1,j)=imagein.pixel(imagein.getWidth()-1,j);
    }
    
    for (int i=0; i<imagein.getWidth(); i++)
    {
    	imageout.pixel(i,0)=imagein.pixel(i,0);
      imageout.pixel(i,imagein.getHeight()-1)=imagein.pixel(i,imagein.getHeight()-1);
    }
}
 
template<typename T>
void despeckleMedian(mm_image<T> &imagein)
{
    mm_image<T> tempimage;

    despeckleMedian(imagein, tempimage);

    imagein = tempimage;
}

template<typename T>
void chamfer(mm_image<T> &inimage)
{
    int nw,n,ne,w,e,sw,s,se;
    int fp;

    for (int j=1; j<inimage.getHeight()-1; j++)
        for (int i=1; i<inimage.getWidth(); i++)
        {
            fp=inimage.pixel(i,j);
            nw=4+inimage.pixel(i-1,j-1);
            n=3+inimage.pixel(i,j-1);
            w=3+inimage.pixel(i-1,j);
            sw=4+inimage.pixel(i-1,j+1);

            fp=fp<nw?fp:nw;
            fp=fp<n?fp:n;
            fp=fp<w?fp:w;
            fp=fp<sw?fp:sw;
            inimage.pixel(i,j)=fp;

        }

    for (int j=inimage.getHeight()-2; j>0; j--)
        for (int i=inimage.getWidth()-2; i>=0; i--)
        {
            fp=inimage.pixel(i,j);
            ne=4+inimage.pixel(i+1,j-1);
            e=3+inimage.pixel(i+1,j);
            s=3+inimage.pixel(i,j+1);
            se=4+inimage.pixel(i+1,j+1);

            fp=fp<ne?fp:ne;
            fp=fp<e?fp:e;
            fp=fp<s?fp:s;
            fp=fp<se?fp:se;
            inimage.pixel(i,j)=fp;

        }
    
    for (int i=0; i<inimage.getWidth(); i++)
    {
      inimage.pixel(i,0)=inimage.pixel(i,1);
      inimage.pixel(i,inimage.getHeight()-1)=inimage.pixel(i,inimage.getHeight()-2);
    }
    
    for (int j=0; j<inimage.getHeight(); j++)
    {
      inimage.pixel(0,j)=inimage.pixel(1,j);
      inimage.pixel(inimage.getWidth()-1,j)=inimage.pixel(inimage.getWidth()-2,j);
    }
 }


template<typename T>
void paethRotateRadian(const mm_image<T> &imagein, mm_image<T> &imageout,
                       float radian)
{
  assert( &imageout != &imagein );
  imageout.init(imagein.getWidth(),imagein.getHeight());
  imageout.fill(255);

    if (fabs(radian-M_PI)<1e-3) radian=radian<M_PI?M_PI-1e-3:M_PI+1e-3; // Quick hack to fix the problem when tan goes to infinity
    float s1=-tan(-radian/2.0);
    float s2=sin(-radian);

    float xr, yr;
    int x,y;
    for (int j=0; j<imageout.getHeight(); j++)
        for (int i=0; i<imageout.getWidth(); i++)
        {
            x=i-imageout.getWidth()/2;
            y=j-imageout.getHeight()/2;
            //x=i;
            //y=j;

            xr = (float)x + s1 * (float)y;
            yr = (float)y + s2 * xr;
            xr = xr + s1 * yr;

            xr+=(float)imageout.getWidth()/2;
            yr+=(float)imageout.getHeight()/2;
            imageout.pixel(i, j)=imagein.getPixel((int)xr,(int)yr);
            //imageout.setPixel(i,j, imagein.getPixel(i,j));
        }
}

template<typename T>
void multiLevelCorrelation(mm_image<T> &edgemap, mm_image<T> &objectedgemap, int
                           levels, int &x, int &y, float &corrval, bool
			   chamferObject)
{
    mm_image<T> object[levels], image[levels];
    
    image[0] = edgemap;
   // invert(image[0]);

    object[0] = objectedgemap;
   // invert(object[0]);
  
     for (int l=1; l<levels; l++)
    {
        halfResolutionEdgePyramid(image[l-1],image[l]);
        halfResolutionEdgePyramid(object[l-1],object[l]);
    }
    
    for (int l=0; l<levels; l++)
    {
      image[l] = (T)255 - image[l];
      object[l] = (T)255 - object[l];
	//image[l].display();
	//object[l].display();
	chamfer(image[l]);
	if (chamferObject) chamfer(object[l]);
    }
 
     int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=image[levels-1].getWidth();
    ymaxsearch=image[levels-1].getHeight();

    for (int l=levels-1; l>=0; l--)
    {
	//object[l].display();
	//image[l].display();

        mincorr=FLT_MAX;
        for (int j=yminsearch; j<ymaxsearch-object[l].getHeight(); j++)
            for (int i=xminsearch; i<xmaxsearch-object[l].getWidth(); i++)
            {
                corr=0.0;
		int count=0;
                for (int b=0; b<object[l].getHeight(); b++)
                    for (int a=0; a<object[l].getWidth(); a++)
                    {
                        if (!chamferObject) if (object[l].pixel(a,b)>127) continue;
                        corr+=(float)abs((int)image[l].pixel(i+a,j+b)-(int)object[l].pixel(a,b));
			count++;
                    }

		corr/=count;
                if (corr<mincorr)
                {
                    xmin=i; ymin=j;
                    mincorr=corr;
                }
            }

        if (l!=0)
        {
            xminsearch=(xmin*2)-1-20;
            yminsearch=(ymin*2)-1-20;

            xmaxsearch=(xmin*2)-1+object[l-1].getWidth()+20;
            ymaxsearch=(ymin*2)-1+object[l-1].getHeight()+20;

            if (xminsearch<0) xminsearch=0;
            if (yminsearch<0) yminsearch=0;

            if (xmaxsearch>image[l-1].getWidth()) xmaxsearch=image[l-1].getWidth();
            if (ymaxsearch>image[l-1].getHeight()) ymaxsearch=image[l-1].getHeight();
        }


    }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void correlation(const mm_image<T> &imagein, mm_image<T> &templatein,
                 int &x, int &y, float &corrval)
{

    int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=imagein.getWidth();
    ymaxsearch=imagein.getHeight();

    mincorr=FLT_MAX;
    for (int j=yminsearch; j<ymaxsearch-templatein.getHeight(); j++)
        for (int i=xminsearch; i<xmaxsearch-templatein.getWidth(); i++)
        {
            corr=0.0;
            for (int b=0; b<templatein.getHeight(); b++)
                for (int a=0; a<templatein.getWidth(); a++)
                {
                    //if (object[l].getPixel(a,b)>127) continue;
                    corr+=(float)abs((int)imagein.pixel(i+a,j+b)-(int)templatein.pixel(a,b));
                }

            if (corr<mincorr)
            {
                xmin=i; ymin=j;
                mincorr=corr;
            }
        }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void correlationChamfer(const mm_image<T> &imagein, mm_image<T> &templatein,
                        int &x, int &y, float &corrval)
{

    int xmin = INT_MIN, ymin = INT_MIN;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float corr, mincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=imagein.getWidth();
    ymaxsearch=imagein.getHeight();

    mincorr=FLT_MAX;
    for (int j=yminsearch; j<ymaxsearch-templatein.getHeight(); j++)
        for (int i=xminsearch; i<xmaxsearch-templatein.getWidth(); i++)
        {
            corr=0.0;
            for (int b=0; b<templatein.getHeight(); b++)
                for (int a=0; a<templatein.getWidth(); a++)
                {
		    if (templatein.pixel(a,b)>127) continue;
                    corr+=(float)abs((int)imagein.pixel(i+a,j+b)-(int)templatein.pixel(a,b));
                }

            if (corr<mincorr)
            {
                xmin=i; ymin=j;
                mincorr=corr;
            }
        }

    assert( xmin != INT_MIN && ymin != INT_MIN );
    x=xmin;
    y=ymin;
    corrval=mincorr;

}

template<typename T>
void rotationChamferCorrelationEdgeMap(mm_image<T> &edgemap, mm_image<T>
	&objectedgemap, int angle1, int angle2, int anglestep, int &x, int &y, int &outangle, float &corrval)
{
    mm_image<T> chamferedImage, rotatedObject, chamferedObject;

    chamferedImage = (T)255 - edgemap;
    chamfer(chamferedImage);

  
    objectedgemap = (T)255 - objectedgemap;
    // Make sure that object boundaries are white:
    drawLine(objectedgemap, 0,0,0,objectedgemap.getHeight()-1,255);
    drawLine(objectedgemap, objectedgemap.getWidth()-1,0,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    drawLine(objectedgemap, 0,0,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    drawLine(objectedgemap, 0,objectedgemap.getHeight()-1,objectedgemap.getWidth()-1,objectedgemap.getHeight()-1,255);
    //chamfer(objectedgemap);

    float cval;
    int tx,ty;
    corrval=FLT_MAX;
    for (int angle=angle1; angle<=angle2; angle+=anglestep)
    {
   rotatedObject = objectedgemap;
	//invert(rotatedObject);
	paethRotateRadian(rotatedObject,chamferedObject,(float)angle*M_PI/180.0);
	//chamfer(chamferedObject);
	correlationChamfer(chamferedImage, chamferedObject, tx, ty, cval);

	if (cval<corrval)
	{
	    corrval=cval;
	    x=tx;
	    y=ty;
	    outangle=angle;
	}
    }
}

template< typename T >
mm_image< T > focusEnhance( const mm_image< T > &image )
{
  const T enhance[3][3]={ { -1,  0, -1 },
                          {  0,  7,  0 },
                          { -1,  0, -1 } };
  boost::const_multi_array_ref< T, 2 > filter( &enhance[0][0],
                                               boost::extents[3][3] );
  return
    normalise( mm_image< T >( correlate( image.rawData(), filter ) ),
               (T)0, (T)255 );
}

template< typename T >
mm_image< T > featureEnhance( const mm_image< T > &image )
{
  const T enhance[3][3]={ {  0, -1,  0 },
                          { -1, 10, -1 },
                          {  0, -1,  0 } };
  boost::const_multi_array_ref< T, 2 > filter( &enhance[0][0],
                                               boost::extents[3][3] );
  return
    normalise( mm_image< T >( correlate( image.rawData(), filter ) ),
               (T)0, (T)255 );
}

template< typename T >
mm_image< T > emboss( const mm_image< T > &image )
{
  const T diff[3][3]={ { -1,  0,  0 },
                       {  0,  0,  0 },
                       {  0,  0,  1 } };
  boost::const_multi_array_ref< T, 2 > filter( &diff[0][0],
                                               boost::extents[3][3] );
  return
    normalise( mm_image< T >( correlate( image.rawData(), filter ) ),
               (T)0, (T)255 );
}

template< typename T >
mm_image< T > softenHeavy( const mm_image< T > &image )
{
  const T median[3]={ 1, 1, 1 };
  boost::const_multi_array_ref< T, 2 > filterx( &median[0],
                                                boost::extents[1][3] );
  boost::const_multi_array_ref< T, 2 > filtery( &median[0],
                                                boost::extents[3][1] );
  return
    normalise( mm_image< T >( correlate( correlate( image.rawData(),
                                                    filterx ), filtery ) ),
               (T)0, (T)255 );
}

template< typename T >
mm_image< T > softenMedium( const mm_image< T > &image )
{
  const T soften[3][3]={ {  1,  1,  1 },
                         {  1,  2,  1 },
                         {  1,  1,  1 } };
  boost::const_multi_array_ref< T, 2 > filter( &soften[0][0],
                                               boost::extents[3][3] );
  return
    normalise( mm_image< T >( correlate( image.rawData(), filter ) ),
               (T)0, (T)255 );
}

template< typename T >
mm_image< T > softenLight( const mm_image< T > &image )
{
  const T soften[3][3]={ {  6,  12,  6 },
                         { 12,  25, 12 },
                         {  6,  12,  6 } };
  boost::const_multi_array_ref< T, 2 > filter( &soften[0][0],
                                               boost::extents[3][3] );
  return
    normalise( mm_image< T >( correlate( image.rawData(), filter ) ),
               (T)0, (T)255 );
}

template<typename T>
void oilPainting(const mm_image<T> &imagein, mm_image<T> &imageout,
                 int regiondim)
{
   const int fy=regiondim;
    const int fx=regiondim;

    assert( &imageout != &imagein );
    imageout.init(imagein.getWidth(),imagein.getHeight());
    
    for (int y=0; y<imagein.getHeight(); y++)
	for (int x=0; x<imagein.getWidth(); x++)
	{
	    if (y-fy/2<0 || y+fy/2>imagein.getHeight()-1 || x-fx/2<0 || x+fx/2>imagein.getWidth()-1)
	    {
         imageout.pixel(x,y)=imagein.pixel(x,y);
		continue;
	    }

       std::vector<int> pixcolor(regiondim*regiondim);
       std::vector<int> graylevel(256);
    
	    int count=0;
	    for (int j=-fy/2; j<=fy/2; j++)
		for (int i=-fx/2; i<=fx/2; i++)
		{
		    pixcolor[count]=imagein.pixel(x+i,y+j);
		    graylevel[pixcolor[count]]+=1;
		    
		    count++;
		}
	    
	    count=0;
	    int intensity=0, maxuse=0;
	    for (int j=-fy/2; j<=fy/2; j++)
		for (int i=-fx/2; i<=fx/2; i++)
		{
		    if (graylevel[pixcolor[count]]>maxuse)
		    {
			maxuse=graylevel[pixcolor[count]];
			intensity=pixcolor[count];
		    }

		    count++;
		}
	    
	    imageout.pixel(x,y)=(T)(intensity);
	}

    
}

template<typename T>
void displayWithOverlay( mm_image_output< mm_rgba< unsigned char > > &display,
                         mm_image<T>& im, mm_property_image& pim)
{
  //alas can't be part of the base class. circular dependencies :-(
  mm_image< mm_rgba< unsigned char > > output;
  output.init( im.getWidth(), im.getHeight() );
    
  for( int j=0;j<im.getHeight();j++)
    for( int i=0;i<im.getWidth();i++) {
      mm_prop_ptr p( pim.pixel(i,j) );
      if( p )
	p->draw( output, i, j, false);
    }
    
  for( int j=0;j<im.getHeight();j++)
    for( int i=0;i<im.getWidth();i++)
      if ( (int)output.pixel( i, j ) ==  0 ) {
	unsigned char value = (unsigned char)im.pixel( i, j );
	output.pixel( i, j ) =
	  mm_rgba< unsigned char >( value, value, value );
      };
          
  display << output;
}

//some common features here with mm_vector class ...
//a cleverer man could probably use the same underlying base class
//but for now it doesnt matter
template< typename T >
double dotProduct( const mm_image<T> &im1,
                   const mm_image<T> &im2 ) throw (mm_exception)
{
  MMERROR( im1.getWidth() == im2.getWidth() &&
           im1.getHeight() == im2.getHeight(),
           mm_exception, ,
           "Can't (won't) compute the dot product of two images of "
           "differing sizes." );

  double result = 0.0;

  const int size = (signed)im1.rawData().num_elements();
  const T
    *p = im1.rawData().data(),
    *q = im2.rawData().data();

  for ( int i=0; i<size; i++ )
    result += *p++ * *q++;

  return result;
}

}

#endif
