#ifndef MM_IMAGE_INPUT_H
#define MM_IMAGE_INPUT_H

#include "mm_exception.h"
#include "mm_image.h"

namespace mimas {

/** @defgroup imageIO Image I/O
    The I/O-facilities for images of mimas are implemented using inheritance.
    This makes it possible to write programs, where the input-source and the
    output-target for an image (or several images) are exchangable.
    @{ */
/** @defgroup imageInput Image Input
    This group contains classes for retrieving image(s) from different input
    sources.
    @{ */
/** Abstract base-class for input sources.
    This is a base-class for input sources. It allows to implement a program,
    which retrieves an image from an arbitrary input-source. F.e.:
    \code
    using namespace mimas;
    ...
    void test( mm_image_input< mm_rgba< unsigned char > > &source ) {
       mm_image< mm_rgba< unsigned char > > image;
       if ( source >> image ) {
         ...
       } else {
         // End of input has been reached.
       }
       ...
    };
    ...
    \endcode */
template< typename T >
class mm_image_input: public mm_object
{
 public:
  /** Constructor.
      The state is set to valid (end of stream not encountered yet). */
  mm_image_input(void): state( true ) {}
  /** Read image from arbitrary source.
      This function is virtual and has to be implemented by the inheriting
      class.
      @param image Object to store image in. */
  virtual void read( mm_image< T > &image ) throw (mm_exception) = 0;
  /// Indicator for input-source being alive.
  operator bool(void) const { return state; }
  /// Indicator for end of input-source.
  bool operator!(void) const { return !state; }
 protected:
  /// State variable.
  bool state;
};

/// Stream-operator for arbitrary input-source.
template< typename T >
inline mm_image_input< T > &operator>>( mm_image_input< T > &input,
                                        mm_image< T > &image )
{ input.read( image ); return input; }

///@}

///@}

};

#endif
