#ifndef MM_IMAGE_QTOUTPUT
#define MM_IMAGE_QTOUTPUT

#include <boost/smart_ptr.hpp>
#include <qlabel.h>
#include "mm_image_output.h"
#include "mm_timer.h"
#include "mm_rgba.h"

namespace mimas {

/** Method calling QApplication::quit.
    Including qapplication.h in this header-file didn't work. */
void quitQtApplication( QApplication *app );

/** @addtogroup imageIO
    @{ */
/** @addtogroup imageOutput
    @{ */
/** Class for displaying image(s) in a window.
    The images are displayed in a window. This class is using the
    <A HREF="http://www.trolltech.com">Qt-library</A>.

    It is possible to use several windows at the same time. If several
    objects are feed with images in a loop, specifying different framerates
    using the \c delay method won't work properly. */
template< typename T >
class mm_image_qtoutput: public mimas::mm_image_output< T >
{
 public:
  /// Constructor.
   mm_image_qtoutput( QApplication &_qApp );
  /** Display image in window.
      If the window was closed, it will be reopened.
      @image Image to be displayed. */
  virtual void write( const mm_image< T > &image );
  /** Wait the specified amount of time.
      Use this method to wait, when windows are displayed, because this
      method will allow the Qt-library to process update events issued by
      X11.
      @param milliSeconds Number of milliseconds to wait. 
      @see delay */
  void wait( int milliSeconds );
  /** Maintain minimum timespan between display of two consecutive frames.
      This will cause the display of the next image to be delayed, so that
      the timespan between displaying of to images will be at least the
      specified amount of time.
      @param milliSeconds Minimum timespan between displaying of two
      onsecutive frames. */
  void delay( int milliSeconds ) { interval = milliSeconds; }
  /** Pause until window is closed by user.
      The method will not return, until the user has closed the window.
      X11 update events will be processed in a loop.
      If the window was closed, it will be reopened. */
  void pause(void);
 protected:
  /// Timer class for internal use.
  class mm_timing: public QObject {
  public:
    /** Constructor starting timer.
        @param _qApp Pointer to application object.
        @param milliSeconds Delay for timer. */
    mm_timing( QApplication *_qApp, int milliSeconds ): qApp(_qApp)
       { startTimer( milliSeconds ); }
  protected:
    /** Event method to be called.
        This method will get called by the Qt-framework. This allows to
        call the method \c QApplication::exec and regain control after a
        specific amount of time.

        This method will call \c QApplication::quit */
        virtual void timerEvent( QTimerEvent * ) { quitQtApplication( qApp ); }
    /// Pointer to application object.
    QApplication *qApp;
  };
  /// Write a pixmap to the window.
  void writePixmap( const QPixmap &pix );
  /// Pointer to application object.
  QApplication *qApp;
  /// Qt-widget for displaying images.
  boost::shared_ptr< QLabel > label;
  /** Timer object for measuring the time.
      This object is used for providing the functionality offered by the
      method \c delay.
      @see delay 
      @see mm_timer */
  mm_timer time;
  ///
  int interval;
  ///
  int lag;
  /** Indicator, wether the window has been opened by the program already.
      This indicator is used for deducing, wether the user has closed the
      window. */
  bool windowWasOpen;
};

template<>
void mm_image_qtoutput< unsigned char >::write
   ( const mm_image< unsigned char > &image );

template<>
void mm_image_qtoutput< mm_rgba< unsigned char > >::write
   ( const mm_image< mm_rgba< unsigned char > > &image );

/// Stream-operator for displaying images.
template< typename T >
inline mm_image_qtoutput< T > &operator<<
( mm_image_qtoutput< T > &output, const mm_image< T > &image )
{ output.write( image ); return output; }

/// Stream-operator for waiting.
template< typename T >
inline mm_image_qtoutput< T > &operator<<
( mm_image_qtoutput< T > &output, _wait wait )
{ output.wait( wait.milliSeconds ); return output; }

/// Stream-operator for adjusting frame-rate.
template< typename T >
inline mm_image_qtoutput< T > &operator<<
( mm_image_qtoutput< T > &output, _delay delay )
{ output.delay( delay.milliSeconds ); return output; }

/// Stream-operator for pausing.
template< typename T >
inline mm_image_qtoutput< T > &operator<<
( mm_image_qtoutput< T > &output, _pause )
{ output.pause(); return output; }

/// For convencience.
inline mm_image_qtoutput< unsigned char > &operator<<
( mm_image_qtoutput< unsigned char > &output,
  const mm_image< int > &image )
{
  output.write( mm_image< unsigned char >( image ) );
  return output;
}

/// For convencience.
inline mm_image_qtoutput< unsigned char > &operator<<
( mm_image_qtoutput< unsigned char > &output,
  const mm_image< double > &image )
{
  output.write(mm_image< unsigned char >( image ) );
  return output;
}

/// For convencience.
inline mm_image_qtoutput< mm_rgba< unsigned char > > &operator<<
( mm_image_qtoutput< mm_rgba< unsigned char > > &output,
  const mm_image< mm_rgba< int > > &image )
{
  output.write( mm_image< mm_rgba< unsigned char > >( image ) );
  return output;
}

/// For convencience.
inline mm_image_qtoutput< mm_rgba< unsigned char > > &operator<<
( mm_image_qtoutput< mm_rgba< unsigned char > > &output,
  const mm_image< mm_rgba< float > > &image )
{
  output.write( mm_image< mm_rgba< unsigned char > >( image ) );
  return output;
}

/// For convencience.
inline mm_image_qtoutput< mm_rgba< unsigned char > > &operator<<
( mm_image_qtoutput< mm_rgba< unsigned char > > &output,
  const mm_image< mm_rgba< double > > &image )
{
  output.write( mm_image< mm_rgba< unsigned char > >( image ) );
  return output;
}

///@}

///@}

};

#include "mm_image_qtoutput.tcc"

#endif
