#ifndef MM_MESA_OFFSCREEN_H
#define MM_MESA_OFFSCREEN_H

#include <boost/shared_ptr.hpp>
#include "mm_image.h"
#include "mm_mesa_base.h"
#include "mm_rgba.h"

#ifndef MMUSEOSMESA
#define GLX_GLXEXT_PROTOTYPES
#include <GL/glx.h>
#else
#include <GL/osmesa.h>
#endif

namespace mimas {

/** Mesa off-screen rendering.
    Support for off-screen rendering using Mesa.

    No stencil- and accumulation-buffer are allocated.
    
    References:
    \li <A HREF="http://www.mesa3d.org/brianp/sig97/offscrn.htm">http://www.mesa3d.org/brianp/sig97/offscrn.htm</A> */
template< typename T >
class mm_mesa_offscreen: public mm_mesa_base
{
 public:
  /** Constructor.
      @param _display X11 display
      @param image Image to draw on.
      @param depthBuffer Boolean indicating, wether to use a depth-buffer. */
  mm_mesa_offscreen( mm_x11_display_ptr _display, mm_image< T > &image,
                     bool depthBuffer = false );
  /** Destructor.
      The destructor just calls the method \c finish() .
      @see finish */
  virtual ~mm_mesa_offscreen(void);
  /** Finish output on the image.
      This method
      \li calls \c glFinish()
      \li copies back the image-data, if \c MMUSEOSMESA is not defined.
      \li destructs all the resources related to off-screen rendering.
      \li finally calls \c mm_mesa_base::finish .
      @see mm_mesa_base::finish */
  virtual void finish(void);
  /** Switch to two-dimensional drawing mode.
      @see mm_mesa_base::ortho2D */
  virtual void ortho2D(void) const;
 protected:
  ///
  GLenum getFormat(void) const;
  ///
  void setupColorScale(void) const;
#ifndef MMUSEOSMESA
  /// X11 pixmap.
  Pixmap pixmap;
  /// Corresponding GLX pixmap
  GLXPixmap glxPixmap;
  /// GLX context for off-screen rendering.
  GLXContext context;
  /// Pointer to image-data to write back later.
  void *data;
#else
  /// Context for (direct) off-screen rendering using OSMesa.
  OSMesaContext context;
#endif
  /// Width of image.
  int width;
  /// Height of image.
  int height;
  ///
  bool finishOffscreen;
};

};

#include "mm_mesa_offscreen.tcc"

#endif
