//
// Multi-layered perceptron tool class
//


#ifndef MM_NN_MLP_TOOL_H
#define MM_NN_MLP_TOOL_H

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/shared_ptr.hpp>
#include <iostream>
#include <stdarg.h>
#include <math.h>
#include "mm_object.h"
#include "mm_exception.h"

namespace mimas {


/**
 * Multi-layered perceptron (MLP) tool
 * @author Bala Amavasai (bala@amavasai.org)
 * @date Tue Sep 24 13:19:08 BST 2002
 * $Header: /cvs/mimas/include/mm_nn_mlp_tool.h,v 1.13 2005/04/28 22:15:28 engjw Exp $
 */


class mm_nn_mlp_tool : public mm_object
{
   public:
      typedef boost::numeric::ublas::matrix< double > Matrix;
      typedef boost::numeric::ublas::vector< double > Vector;
	private:
		typedef double (mm_nn_mlp_tool::*ptrToFuncType)(double); ///< pointer-to-member-function type
      boost::shared_array< ptrToFuncType > ptrToTransferFunc;
		boost::shared_array< ptrToFuncType > ptrToTransferFuncDeriv;
      boost::shared_array< Matrix > weight; ///< the weights
		boost::shared_array< Vector > bias; ///< the biases
		boost::shared_array< Vector > y; ///< outputs of nodes (after transfer function)
		boost::shared_array< Vector > net; ///< the net output of nodes (before transfer function)
		boost::shared_array< Vector > delta; ///< the delta values from the delta rule
		boost::shared_array< Matrix > prevWeightChange; ///< the previous weight updates
		boost::shared_array< Vector > prevBiasChange; ///< the previous bias update
      boost::shared_array< int > numnodes; ///< the number of nodes in each layer
		bool network_initialised; ///< is the network initialised?
		bool debugtrain; ///< show debugging option when training?
		int numlayers; ///< this include the input and output layers
		//double learningRate, momentum; ///< learning rate and momentum
		int iterations; ///< number of iterations to train for
		void init(void); ///< private function which initialises network based on numnodes[]
		double deltaRule( Matrix& pat, int p) throw (mm_exception); // perform delta rule on row p of pattern matrix pat

	public:
		mm_nn_mlp_tool(void); ///< constructor
		~mm_nn_mlp_tool(void); ///< destructor
		void init(int nodes, ...) throw (mm_exception); ///< initialise network with number of nodes in each layer. List must be NULL terminated!
		void clear(void); ///< delete the whole network
		void showConfig(void); ///< show network configuration
      Matrix feedForward( Matrix& pat) throw (mm_exception); ///< feedforward patterns and return a matrix containing the results.

		// io functions
		void save (const char *fn) throw (mm_exception); ///< save neural network weights to a file. Note that transfer function information are not saved!
		void load (const char *fn) throw (mm_exception); ///< load neural network from a file. Note that transfer function information are not loaded!


		// accessor functions
		void setIterations(int i) throw (mm_exception); ///< set number of iterations to train for
		int getIterations(void); ///< get number of iterations to train for
		
		// training methods
      Matrix trainOnline( Matrix& pat, double learningRate=0.3, double momentum=0.05, double weightDecay=0.0) throw (mm_exception); ///< online training of network on input/output dataset pat. Note that the number of columns that are the input patters and target patters are determined by the number of inputs and outputs of the network during the init() call. Returns 2-column matrix of sum-squared-error (SSE) values.
      Matrix trainBatch(Matrix& pat, double learningRate=0.3, double momentum=0.05, double weightDecay=0.0) throw (mm_exception); ///< batch training of network on input/output dataset pat. Note that the number of columns that are the input patters and target patters are determined by the number of inputs and outputs of the network during the init() call. Returns 2-column matrix of sum-squared-error (SSE) values.
		Matrix trainQprop(Matrix& pat, double learningRate=0.0005, double momentum=0.05, double maxFactor=2.0) throw (mm_exception); ///< batch training using Quickprop on input/output dataset pat. Note that the number of columns that are the input patters and target patters are determined by the number of inputs and outputs of the network during the init() call. Returns 2-column matrix of sum-squared-error (SSE) values.
      Matrix trainRprop(Matrix& pat, double etapos=1.2, double etaneg=0.5) throw (mm_exception); ///< batch training using RProp on input/output dataset pat. Note that the number of columns that are the input patters and target patters are determined by the number of inputs and outputs of the network during the init() call. Returns 2-column matrix of sum-squared-error (SSE) values.

		// set transfer function for a layers
		void setTransferFuncLayer(ptrToFuncType newfunc, int layer) throw (mm_exception); ///< set transfer function for layer. Note that first hidden layer is layer 2. Must be set together with setTransferFuncDerivLayer()
		void setTransferFuncDerivLayer(ptrToFuncType newfunc, int layer) throw (mm_exception); ///< set derivative of transfer function for layer. Note that first hidden layer is layer 2. Must be set together with setTransferFuncLayer()
		
		// user selectable transfer functions and the derivatives of the functions
		virtual double transferFunc(double val); ///< a transfer function that is overwritable by the user
		virtual double transferFuncDeriv(double val); ///< the derivative of a transfer function that is overwritable by the user
		double transferFuncTanh(double val); ///< the hyperbolic tangent transfer function
		double transferFuncDerivTanh(double val); ///< the derivative of hyperbolic tangent the transfer function
		double transferFuncLinear(double val); ///< the linear transfer function
		double transferFuncDerivLinear(double val); ///< the derivative of linear the transfer function

		// debugging options
		void setDebug(bool val); ///< set/unset debugging flag. When set, training informaiton will be displayed
	
};

}
#endif


