#ifndef MM_PGH_2DOBJECT
#define MM_PGH_2DOBJECT

#include <boost/numeric/ublas/vector.hpp>
#include <boost/smart_ptr.hpp>
#include <iostream>
#include <list>
#include <string>
#include "mm_config.h"
#include "mm_exception.h"
#include "mm_pgh.h"
#include "mm_object.h"
#include "mm_pgh_type.h"
#include "mm_image.h"
#ifdef HAVE_XALAN
#include "mm_xml_reference.h"
#endif

namespace mimas {
/** Pairwise model representation.
    Flat 2d objects with pairwise these objects store info as a series of
    pgh's. A representation of a 2d model. this is a list of PGHs, one created
    for each line in the original model. also included is the original line
    information, primarily for rendering purposes.
   
    OK a real object is a set of histograms which describe a set of real lines.
    That's what this class is for. Eg. \c mm_pgh_2dobject my_dinoraur_template.
    Use it basically by doing the following:
    \code
    mm_pgh_2dobject	my_dino_template;
    my_dino_template.createFromLines( set_of_lines ); 
    my_dino_template.setName("Arthur_the_stegosaurus");
    \endcode

    @author Stuart Meikle (stu@stumeikle.org)
    @date Thu May 11 13:05:23 2000

    Added new \c addModel function
    \code createFromPgmFile(char *pgmfilename , const char *modelname )\endcode
    @author bpa
    @date Tue Apr 17 13:49:36 BST 2001 */
class	mm_pgh_2dobject : public mm_object
{
 public:
  typedef std::string string;
  typedef boost::numeric::ublas::vector< double > Vector;
private:
	string			name;  ///< object name, used for referencs
	mm_pgh_ptr_list pghlist; ///< the list of pghs which form the model
	int			num_pghs;///< number of pghs in the model


	//co-ordinate system somehow --> we have the problem that there
	//could be many instances of the object on the image scene

	//other info:
   mm_line_ptr_list_ptr poly_data;///< straight lines which form the (input) model.

	//for display purps
	//translated to be centred on the origin
	int	bounding_box_x, bounding_box_y;///< range of the straight lines on the image plane


private:
	void		deletePghList(void);


protected:
	void		deleteLineInfo(void);

public:
	// added by bpa - for non centered lines
	mm_line_ptr_list_ptr poly_data_GUI;

	// Canny parameters, added by fbc, 11.10.2001

	/**
	 * models can be formed from images. in this case edge strings must be
	 * extracted first. the following parameters specify the parameters which
	 * are passed to the canny detector for this first step.
	 */
	double		canny_sigma;	///< see mm_canny_tool.h
	double		canny_precision;///< see mm_canny_tool.h
	double		canny_lowthre;///< see mm_canny_tool.h
	double		canny_upthre;///< see mm_canny_tool.h
	int		canny_lenthre;///< see mm_canny_tool.h

	// ROI parameters, added by fbc, 16.10.2001
	int			roi_lx;///< region of interest in the incoming image, lower x
	int			roi_ly;///< region of interest in the incoming image, lower y
	int			roi_ux;///< region of interest in the incoming image, upper x
	int			roi_uy;///< region of interest in the incoming image, upper y

	// these should be derived from the mm_pgh class...I think, fbc 11.10.2001
	/**
	 * parameters used during the creation of the 2bobjects constituent pghs
	 * passed to mm_pgh.
	 */
	mm_pgh_type pgh_pairs_type;///< see mm_pgh.h
	double		pgh_dbin_max;///< see mm_pgh.h
	double		pgh_dbin_size;///< see mm_pgh.h
	double		pgh_dist_ramp;///< see mm_pgh.h
	int		pgh_num_abin;///< see mm_pgh.h
	double 		pgh_angle_sigma;///< see mm_pgh.h

	// minimum line length
	double		min_line_length;///< lines under this length will not be considered

	Vector	ref_point_displ;///< object reference point (displacement from its centroid), added by fbc 5.12.2001

	int		track_roi_lx;///<object tracking region (displacement from its ref. point), lower x
	int		track_roi_ly;///<object tracking region (displacement from its ref. point), lower y
	int		track_roi_ux;///<object tracking region (displacement from its ref. point), upper x
	int		track_roi_uy;///<object tracking region (displacement from its ref. point), upper y

   /// The constructor.
   mm_pgh_2dobject(void) { init(); };
	~mm_pgh_2dobject(void);

  /// Set variables to default values (only called by constructor).
   void        init(void);

	void			clear(void);	// added by fbc, 14.01.02

	string	getName(void);
	int			getNumPghs(void);
	mm_pgh_type		getPghType(void);
	void			setName(const std::string &n);
	mm_pgh_ptr_list       *getPghList(void);
   mm_line_ptr_list_ptr getListofLines(void);
	mm_line_ptr_list_ptr getListofLinesGUI(void);
	void			createFromLines( const mm_line_ptr_list &geom ) throw (mm_exception);
	void			createFromLines( const mm_line_ptr_list &geom, double *cx, double *cy) throw (mm_exception);

	// added by bpa
	void			createFromPgmFile(char *pgmfilename);
	void			createFromImage(mm_image<int> &imagein);
//	void 			createFromStructCoord(struct co_ord *obj_out_line, int numcoords);

	// added by fbc, 17.12.01
	void			createFromPolyFile(char *polyfilename);
	void			createFromPolyFile(char *polyfilename, double *cx, double *cy);
	void			createFromPgmFile(char *pgmfilename, double *cx, double *cy);
	void			createFromPgmFile(char *pgmfilename, double *cx, double *cy, char *polyfilename);
	void			createFromImage(mm_image<int> &imagein, double *cx, double *cy);
	void			createFromImage(mm_image<int> &imagein, double *cx, double *cy, char *polyfilename);

	//add: load and save!

	int			getWidth(void)	{ return bounding_box_x; }
	int			getHeight(void) { return bounding_box_y; }
	// void			showAll(void);
#ifdef HAVE_XALAN
	// save parameters to xml-file. Returns true on success, added by Jan Wedekind
	void			saveObjectParameters( const char *fileName );

	// load parameters from xml-file. Returns true on success, added by Jan Wedekind
	void			loadObjectParameters( const char *fileName );

 protected:
	void			storeObjectParams( std::ostream &stream );
	void			extractObjectParams( mm_xml_reference< XalanElement >
                                    &docElement );
#endif
};

typedef boost::shared_ptr< mm_pgh_2dobject > mm_pgh_2dobject_ptr;

typedef std::list< mm_pgh_2dobject_ptr > mm_pgh_2dobject_ptr_list;

}

#endif
