#ifndef MM_SVD_CORRESPONDENCE_H
#define MM_SVD_CORRESPONDENCE_H

#include <boost/numeric/ublas/matrix.hpp>
#include "mm_object.h"

namespace mimas {
/** @defgroup featureTracking Tracking of Features
    Mimas-API for tracking features.
    @{ */
/** Correspondence by Singular Value Decomposition.
    This class implements the core algorithm described in
    <A HREF="http://www.hpl.hp.com/techreports/97/HPL-97-96.pdf">Maurizio
    Pilu's article</A>.
    
    Given
    \li a map \f$m:I\mapsto F\f$ where \f$I\f$ is an index set and \f$F\f$ is
    a set of features
    \li a second set of features \f$F^\prime\f$.
    \li a similarity measure
    \f$F\times F^\prime\mapsto [0,1]\f$ giving real numbers between zero and
    one

    the algorithm will guess about correspondences and yield a new map
    \f$m^\prime:I^\prime\mapsto F^\prime\f$, where \f$I^\prime\subset I\f$. If
    the algorithm detects ambiguity, it will drop some indices and
    \f$I^\prime\f$ will be a true subset of \f$I\f$.

    \c InputIter1 must implement the concept of
    <A HREF="http://www.sgi.com/tech/stl/InputIterator.html">Input
    Iterator</A> and it must be an iterator on a
    <A HREF="http://www.sgi.com/tech/stl/pair.html">pair</A> with
    an index as first and a feature as second element. Usually this will
    be something like \code map< Index, F >::const_iterator\endcode, where
    \c F denotes the feature class.

    \c SimilarityMeasure must be a 
    <A HREF="http://www.sgi.com/tech/stl/BinaryFunction.html">binary_function< F, F, double ></A>,
    where \c F denotes the feature class.

    @author Jan Wedekind (jan@wedesoft.de)
    @date Thu May 20 15:53:37 UTC 2004 */
template< class InputIter1, class SimilarityMeasure >
class mm_svd_correspondence: public mm_object
{
 public:
  ///
  typedef boost::numeric::ublas::matrix< double, boost::numeric::ublas::column_major > Matrix;
  typedef boost::numeric::ublas::identity_matrix< double > IdentityMatrix;
  /** Constructor.
      The constructor will take the index-map and the similarity measure.
      @param aSimilarityMeasure Similarity measure, giving values
      \f$\in[0,1]\f$.
      @param aBegin1 Iterator on feature-map \f$m\f$.
      @param aSize1 Number of elements in \f$m\f$. */
  mm_svd_correspondence( SimilarityMeasure &aSimilarityMeasure,
                         InputIter1 aBegin1, int aSize1 ):
    similarityMeasure(aSimilarityMeasure), begin1(aBegin1), size1(aSize1) {}
  /** Match against new feature set.
      This method will take a feature set \f$F^\prime\f$ containing new
      features. The method will give the result by writing index-feature-pairs
      to \c output.
      @param begin2 Input iterator on new feature set.
      @param size2 Size of new feature set.
      @param outputIndexed Output iterator to store indexed features in
      (usually an object of type \c map<>::iterator).
      @param outputLost Output iterator to store lost features in. */
  template< class InputIter2, class OutputIter1, class OutputIter2 >
  void match( InputIter2 begin2, int size2, OutputIter1 outputIndexed,
              OutputIter2 outputLost ) throw (mm_exception);
 protected:
  /** Similarity measure.
      An object of type \c binary_function<> giving similarity estimates
      as double values. */
  SimilarityMeasure similarityMeasure;
  /// Iterator on indexed feature set.
  InputIter1 begin1;
  /// Size of indexed feature set.
  int size1;
};

///@}

}

#include "mm_svd_correspondence.tcc"

#endif
