#ifndef __MM_XML_DOCUMENT_H
#define __MM_XML_DOCUMENT_H

#include "mm_config.h"

#ifdef HAVE_XALAN

#include <string>
#include <XalanDOM/XalanDocument.hpp>
#include <XalanDOM/XalanElement.hpp>
#include <XalanSourceTree/XalanSourceTreeDOMSupport.hpp>
#include <XalanSourceTree/XalanSourceTreeParserLiaison.hpp>
#ifndef SIMPLEXPATH
// With which version was this introduced exactly?
#include <XPath/XPathEnvSupportDefault.hpp>
#include <XPath/XObjectFactoryDefault.hpp>
#if _XALAN_VERSION >= 10500
#include <XPath/XPathConstructionContextDefault.hpp>
#endif
#include <XPath/XPathFactoryDefault.hpp>
#include <XPath/XPathExecutionContextDefault.hpp>
#endif
#include <util/XMLException.hpp>
#include "mm_exception.h"
#include "mm_object.h"
#include "mm_xml_errorhandler.h"
#include "mm_xml_reference.h"

#ifdef XERCES_HAS_CPP_NAMESPACE
using namespace XERCES_CPP_NAMESPACE;
#endif
#ifdef XALAN_HAS_CPP_NAMESPACE
using namespace XALAN_CPP_NAMESPACE;
#endif

namespace mimas {

/** @defgroup xml XML parser
    For parsing and verifying XML-documents, the powerful Xerces-C and
    Xalan-C libraries from <A HREF="http://xml.apache.org/">Apache</A> are
    used.
    
    For more information, have a look at the pages hosted by the <B>Apache
    Software Foundation</B> and the <B>W3C-consortium</B>:
    \li <A HREF="http://xml.apache.org/">http://xml.apache.org/</A>
    \li <A HREF="http://www.w3.org/TR/xmlschema-0/">http://www.w3.org/TR/xmlschema-0/</A>
    \li <A HREF="http://www.w3.org/TR/xpath">http://www.w3.org/TR/xpath</A>
    \li <A HREF="http://www.w3.org/TR/xml11/">http://www.w3.org/TR/xml11/</A>
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed Jan 12 13:53:41 UTC 2004 
    @{ */
/** Class for parsing XML documents.
    Don't forget to add the following initialization code to your main
    application, before using this class:
    \verbatim
    XMLPlatformUtils::Initialize();
    XPathInit oTheXPathInit;
   \endverbatim
   Which means that you will have to do the following includes:
   \verbatim
   #include <XPath/XPathInit.hpp>
   #include <util/PlatformUtils.hpp>
   \endverbatim
   Don't forget either to add the corresponding include path compiler options
   that will look like, depending on your system installation:
   \verbatim
   -I/usr/include/xalan/
   -I/usr/include/xerces/
   \endverbatim
   After using this class, execute the following code:
    \verbatim
    XMLPlatformUtils::Terminate();
   \endverbatim
    See Apache Xerces and Xalan documentation for more details.
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed Jan 12 13:53:41 UTC 2004 */
class mm_xml_document: public mm_object
{
public:
  /** Initialise xml-parser.
      A URI for verification consists of two parts:
      \li The URI, which is also mentioned in the xmlns-attribute of the
      XML-document to be used later.
      \li The location (usually a file-name of a local file.
      @param schemaURI URI of XSD-file to be used for verifiaction. Specify
      "" for no verification. */
  mm_xml_document( const std::string &schemaURI );
  /** Load XML-file.
      Load a XML-file and perform verification, if necessary.
      @param fileName Name of file to load. */
  void load( const char *fileName ) throw (mm_exception);
  /** Save XML-file.
      @param fileName Name of file to write XML-document to. */
  void save( const char *fileName ) throw (mm_exception);
  /** Read XML-document from string.
      @param string String to read XML-document from. */
  void fromString( const std::string &string ) throw (mm_exception);
  /** Read XML-document from Xerces-input-source.
      @param inputSource An object of input-source-type. */
  void fromInputSource( InputSource &inputSource ) throw (mm_exception);
  /** Get pointer to XML-DOM.
      @return Pointer to XML-DOM. */
  const XalanDocument *getDocument(void) const
  { return document; }
  /** Get root-element of XML-DOM.
      @return Wrapper-object for root-element of XML-DOM. */
  mm_xml_reference< XalanElement > getDocumentElement(void)
  { return mm_xml_reference< XalanElement >
      ( document->getDocumentElement(), this ); }
#ifndef SIMPLEXPATH
  /** @name Methods for internal use only */
  ///@{
  ///
  XalanSourceTreeDOMSupport *getDOMSupport(void)
  { return &theDOMSupport; }
  ///
  XPathEnvSupportDefault *getEnvSupport(void) { return &theEnvSupport; }
#if _XALAN_VERSION >= 10500
  ///
  XPathConstructionContextDefault *getXPathConstructionContext(void)
  { return &theXPathConstructionContext; }
#endif
  ///
  XPathFactoryDefault *getXPathFactory(void) { return &theXPathFactory; }
  ///
  XPathExecutionContextDefault *getExecutionContext(void)
  { return &theExecutionContext; }
  ///@}
#endif
protected:
  /** @name Xalan classes for parsing and maintaining the XML DOM */
  ///@{
  ///
  XalanSourceTreeDOMSupport theDOMSupport;
  ///
  XalanSourceTreeParserLiaison theLiaison;
  //@}
  /// Error handler for catching parse errors.
  mm_xml_errorhandler xmlErrorHandler;
  /// The XML DOM.
  const XalanDocument *document;
#ifndef SIMPLEXPATH
  /** @name Xalan classes for making XPath queries */
  ///@{
  ///
  XPathEnvSupportDefault theEnvSupport;
  ///
  XObjectFactoryDefault theXObjectFactory;
  // With which version was this introduced exactly?
#if _XALAN_VERSION >= 10500
  ///
  XPathConstructionContextDefault theXPathConstructionContext;
#endif
  ///
  XPathFactoryDefault theXPathFactory;
  ///
  XPathExecutionContextDefault theExecutionContext;
  ///@}
#endif
};

///
std::istream &operator>>( std::istream &stream, mm_xml_document &document );

///@}


};

#endif

#endif
