#include <cassert>
#include <cmath>
#include "mm_binmorph_tool.h"

namespace mimas {

  using namespace std;

  mm_binmorph_tool::mm_binmorph_tool(void)
  {
    structElement.init(3,3);
    for (int i=0; i<structElement.getWidth(); i++)
      for (int j=0; j<structElement.getHeight(); j++)
        structElement.pixel(i,j) = 1;
  }

  mm_binmorph_tool::~mm_binmorph_tool(void)
  {
  }


  int mm_binmorph_tool::structElementOnes(void)
  {
    int numOnes=0;
	
    for (int x=0; x<structElement.getWidth(); x++)
      for (int y=0; y<structElement.getHeight(); y++)
	if (structElement.pixel(x,y)==1) numOnes++;

    return numOnes;
  }


  //mb: Having a circular structured element could be interesting for big values (>=8).
  void mm_binmorph_tool::setSizeStructElement(int size)
  {
    structElement.init(size,size);
    for (int i=0; i<structElement.getWidth(); i++)
      for (int j=0; j<structElement.getHeight(); j++)
	structElement.pixel(i,j) = 1;
  }

  void mm_binmorph_tool::greyToBinary(mm_image<int> &imagein, mm_image<int> &imageout, int threslevel)
  {	
    imageout.init(imagein.getWidth(),imagein.getHeight());

    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
        imageout.pixel(i,j) = imagein.pixel(i,j) >= threslevel ? 1 : 0;
  }



  void mm_binmorph_tool::greyToBinaryOtsu(mm_image<int> &imagein, mm_image<int> &imageout)
  {	
	
    MMERROR( imagein.initialised(), mm_exception, ,
	     "mm_binmorph_tool::greyToBinary() - image not initialised" );

    imageout.init(imagein.getWidth(),imagein.getHeight());


    int lhist=256;
    float hist[lhist], hs[lhist];
	
    // reset histogram
    for (int i=0; i<lhist; i++)
      {
	hist[i]=0;
	hs[i]=0;
      }

    // build histogram
    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++) {
	const int value = imagein.pixel( i, j );
	if ((value>=0) && (value<lhist))
	  hist[value]+=1.0;
      };

    // normalise histogram
    float imgsize=(imagein.getWidth()*imagein.getHeight());
    for (int i=0; i<lhist; i++)
      hist[i]=hist[i]/imgsize;

    // calculate image mean 
    float imgmean=0.0;
    for (int i=0; i<lhist; i++)
      imgmean+=(i+1)*hist[i];

    // calculate image variance
    float imgvariance=0.0;
    for (int i=0; i<lhist; i++)
      imgvariance+= (i+1-imgmean)*(i+1-imgmean)*hist[i];


    float omega=0.0;
    float levelmean=0.0;
    float mu=0.0;
    // float ratiovariance=0.0;
    int threshold=0;
    // float largestratio=-1.0;
    float largestmu=-1.0;
	
    for (int i=0; i<lhist; i++)
      {
	omega=0.0;
	levelmean=0.0;
	for (int j=0; j<i; j++)
	  {
	    omega+=hist[j];
	    levelmean+=(j+1)*hist[j];
	  }

	mu=(imgmean*omega-levelmean)*(imgmean*omega-levelmean); // (mu0 * mu1)^2 without denominator
	omega=omega*(1.0-omega); // omega0 * omega1

	if (omega>0.0)
	  mu=mu/omega;
	else
	  mu=0.0;

	/* if variance ration is required 
	   ratiovariance=mu/imgvariance;
		
	   if (ratiovariance>largestratio)
	   {
	   largestratio=ratiovariance;
	   threshold=i;
	   }*/

	// skip redundant computations above and get straight to the point!
	if (mu>largestmu)
	  {
	    largestmu=mu;
	    threshold=i;
	  }
		
      }
		
			
		
    //cerr << "Otsu: " << threshold << endl << endl;

    greyToBinary(imagein,imageout,threshold);
	
	
  }


  void mm_binmorph_tool::greyToBinary(mm_image<int> &imagein, int threslevel)
  {
    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
        imagein.pixel(i,j) = imagein.pixel(i,j)>=threslevel?1:0;
  }


  /** Take 2 images of the same size. If for a given pixel position
      both images have values different from one then the result
      image has a one on these pixel position.
  */ //mb
  mm_image<int> mm_binmorph_tool::image_and(mm_image<int> &image1, mm_image<int> &image2)
  {
    MMERROR( image1.getWidth() == image2.getWidth() &&
	     image1.getHeight() == image2.getHeight(), mm_exception, ,
	     "mm_binmorph_tool::image_and() - input images do not have same "
	     "dimensions" );
    
    mm_image<int> result;
    result.init(image1.getWidth(), image1.getHeight());
    
    for (int i=0; i<image1.getWidth(); i++)
      for (int j=0; j<image1.getHeight(); j++)
	if((image1.pixel(i,j) != 0) && (image2.pixel(i,j) != 0))
	  result.pixel(i,j) = 1;

    return result;
  }

  void mm_binmorph_tool::normaliseToGrey(mm_image<int> &imagein)
  {
    int maxval=MININT;
    int minval=MAXINT;

    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
	{
	  int pixval=imagein.pixel(i,j);
	  maxval=pixval>maxval?pixval:maxval;
	  minval=pixval<minval?pixval:minval;
	}


    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
        imagein.pixel(i,j) = (maxval-minval)!=0?((imagein.pixel(i,j)-minval)*255)/(maxval-minval):0;

  }



  void mm_binmorph_tool::normaliseToColour(mm_image<int> &imagein, mm_image< mm_rgba< int > > &imageout)
  {	
    imageout.init(imagein.getWidth(),imagein.getHeight());
	
    int maxval=MININT;
    int minval=MAXINT;

    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
	{
	  int pixval=imagein.pixel(i,j);
	  maxval=pixval>maxval?pixval:maxval;
	  minval=pixval<minval?pixval:minval;
	}

    MMERROR( maxval != minval, mm_exception, ,
	     "mm_binmorph_tool::normaliseToColour - Cannot normalize plain "
	     "image." );

    int redval, greenval, blueval;
    double mult=255*255*255/(double)(maxval-minval);
	
    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
	{
	  int pixval=(int)(mult*(double)(imagein.pixel(i,j)-minval));
	  redval=pixval%255;
	  greenval=(pixval/255)%255;
	  blueval=(pixval/(255*255))%255;
	  imageout.pixel( i, j ) =
	    mm_rgba< int >( redval, greenval, blueval );
	}
  }


  void mm_binmorph_tool::binaryToGrey(mm_image<int> &imagein)
  {
    normaliseToGrey(imagein);	
  }

  void mm_binmorph_tool::negate(mm_image<int> &imagein)
  {
    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
        imagein.pixel(i,j) = imagein.pixel(i,j)==0 ? 1 : 0;
  }

  void mm_binmorph_tool::morphology(mm_image<int> &imagein, mm_image<int> &imageout, mm_binmorph_tool::operation op )
  {
    int countpix;
    int numones=structElementOnes();

    imageout.init(imagein.getWidth(),imagein.getHeight());

    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
	{
		

	  countpix=0;
	  int pixval=0;
	  for (int x=0; x<structElement.getWidth(); x++)
	    for (int y=0; y<structElement.getHeight(); y++)
	      {
		if ((i+x-structElement.getWidth()/2 >= 0 ) &&
		    (i+x-structElement.getWidth()/2 < imagein.getWidth()) 	&&
		    (j+y-structElement.getHeight()/2 >= 0 ) &&
		    (j+y-structElement.getHeight()/2 < imagein.getHeight()))
		  pixval=imagein.pixel(i+x-structElement.getWidth()/2,j+y-structElement.getHeight()/2);


		if ((structElement.pixel(x,y)==1) && (pixval==1))
		  countpix++;
	      
	      }
	
	  if (op==mm_binmorph_tool::mm_binmorph_erosion) imageout.pixel(i,j) = (countpix==numones)?1:0;
	  if (op==mm_binmorph_tool::mm_binmorph_dilation) imageout.pixel(i,j) = (countpix>0)?1:0;
	}
  }


  void mm_binmorph_tool::erosion(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    morphology(imagein,imageout,mm_binmorph_tool::mm_binmorph_erosion);

  }

  void mm_binmorph_tool::dilation(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    morphology(imagein,imageout,mm_binmorph_tool::mm_binmorph_dilation);
  }

  void mm_binmorph_tool::closing(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    mm_image<int> imagetemp;
	
    morphology(imagein,imagetemp,mm_binmorph_tool::mm_binmorph_dilation);
    morphology(imagetemp,imageout,mm_binmorph_tool::mm_binmorph_erosion);

  }

  void mm_binmorph_tool::opening(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    mm_image<int> imagetemp;
	
    morphology(imagein,imagetemp,mm_binmorph_tool::mm_binmorph_erosion);
    morphology(imagetemp,imageout,mm_binmorph_tool::mm_binmorph_dilation);

  }


  void mm_binmorph_tool::difference(mm_image<int> &imagein1, mm_image<int> &imagein2, mm_image<int> &imageout)
  {

    MMERROR( imagein1.getWidth() == imagein2.getWidth() &&
	     imagein1.getHeight() == imagein2.getHeight(), mm_exception, ,
	     "mm_binmorph_tool::difference() - input imagess do not have same "
	     "dimensions" );

    imageout.init(imagein1.getWidth(),imagein1.getHeight());

    for (int i=0; i<imagein1.getWidth(); i++)
      for (int j=0; j<imagein1.getHeight(); j++)
	{
	  if ((imagein1.pixel(i,j)==1) && (imagein2.pixel(i,j)==0) )
	    imageout.pixel(i,j) = 1;
	  else
	    imageout.pixel(i,j) = 0;
	}
	
  }

  void mm_binmorph_tool::boundaryExtract(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    imageout.init(imagein.getWidth(),imagein.getHeight());
	
    mm_image<int> imagetemp;

    erosion(imagein,imagetemp);
    difference(imagein,imagetemp,imageout);


  }

  void mm_binmorph_tool::labelConnectedGroup(mm_image<int> &imagein, mm_image<int> &imageout, int x, int y, int label)
  {
    // check limits
    if ((x<0) || (x>=imagein.getWidth()) || (y<0) || (y>=imagein.getHeight())) return;

    // end of search
    if ((imagein.getPixel(x,y)!=1) || (imageout.getPixel(x,y)!=0)) return;
	
    imageout.setPixel(x,y,label);

    // note that we are assuming that search starts from top left hand corner
    labelConnectedGroup(imagein, imageout, x-1, y-1, label);
    labelConnectedGroup(imagein, imageout, x, y-1, label);
    labelConnectedGroup(imagein, imageout, x+1, y-1, label);
    labelConnectedGroup(imagein, imageout, x+1, y, label);
    labelConnectedGroup(imagein, imageout, x-1, y+1, label);
    labelConnectedGroup(imagein, imageout, x, y+1, label);
    labelConnectedGroup(imagein, imageout, x+1, y+1, label);
  }


  void mm_binmorph_tool::connectedComponents(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    imageout.init(imagein.getWidth(),imagein.getHeight());

    // initialise imageout to 0
    // for (int i=0; i<imageout.getWidth(); i++)
    // 	for (int j=0; j<imageout.getHeight(); j++)
    // 		imageout.setPixel(i,j,0);

    // start label number with 1
    int label=1;
    for (int i=0; i<imagein.getWidth(); i++)
      for (int j=0; j<imagein.getHeight(); j++)
        if ( (imagein.pixel(i,j)==1) && (imageout.pixel(i,j)==0) )
	  {
	    labelConnectedGroup(imagein,imageout,i,j,label);
	    label++;
	  }

    //cerr << "Found " << label-1 << "labels" << endl;
  }


  // Find the distance transform of a binary image
  // object is 1 and background is 0
  void mm_binmorph_tool::distanceTransform(mm_image<int> &imagein, mm_image<int> &imageout)
  {
    imageout.init(imagein.getWidth(),imagein.getHeight());

    // H0

    int picsize=imageout.getWidth()*imageout.getHeight();
    for (int i=0; i<imageout.getWidth(); i++)
      for (int j=0; j<imageout.getHeight(); j++)
	{
	  imageout.pixel(i,j) = picsize;
	  if (imagein.pixel(i,j)==1)
	    {
	      bool setzero=false;
	      for (int x=-1; x<=1; x++)
		for (int y=-1; y<=1; y++)
		  {
		    if (abs(x)+abs(y) == 0)  continue; // ignore centre pixel
		    if (abs(x)+abs(y) == 2)  continue; // ignore diagonal pixels

		    if ((x+i>=0) && (y+j>=0) && (x+i<imageout.getWidth()) && (y+j<imageout.getHeight()))
		      {
			if (imagein.getPixel(x+i,y+j)==0) setzero=true;
			break;
		      }
		  }

	      if (setzero) imageout.pixel(i,j) = 0;
	    }
	}


    int minval;
    int dir;
    int sqrt2=(int)sqrt(2.0);

    // H1

    //mm_image<int> tempdisp;
    //copy(imageout,tempdisp);
    //normaliseToGrey(tempdisp);
    //display(tempdisp)

    // pass 1
    for (int j=0; j<imageout.getHeight(); j++)
      for (int i=0; i<imageout.getWidth(); i++)
	{
	  if (imageout.pixel(i,j)==0) continue; // point already on the boundary

	  minval=MAXINT;

	  for (int x=-1; x<=1; x++)
	    for (int y=-1; y<=1; y++)
	      {
		if ((x+i<0) || (y+j<0) || (x+i>=imageout.getWidth()) || (y+j>=imageout.getHeight()))
		  continue;

		dir=abs(x)+abs(y);

		int pixval;
		switch ( dir ) {
		case 0:
		  pixval=imageout.getPixel(x+i,y+j); // centre pixel
		  break;
		case 1:
		  pixval=imageout.getPixel(x+i,y+j) + 1; // horizontal and vertical pixels
		  break;
		case 2:
		  pixval=imageout.getPixel(x+i,y+j) + sqrt2; // diagonal pixels
		  break;
		default:
		  pixval = 0;
		  assert( false );
		}

		if (pixval<minval) minval=pixval;

	      }

	  imageout.pixel(i,j) = minval;
	}

    // H2

    // pass 2
    for (int j=imageout.getHeight()-1; j>=0; j--)
      for (int i=imageout.getWidth()-1; i>=0;  i--)
	{
	  if (imageout.pixel(i,j)==0) continue; // point already on the boundary

	  minval=MAXINT;

	  for (int x=-1; x<=1; x++)
	    for (int y=-1; y<=1; y++)
	      {
		if ((x+i<0) || (y+j<0) || (x+i>=imageout.getWidth()) || (y+j>=imageout.getHeight()))
		  continue;

		dir=abs(x)+abs(y);

		int pixval;
		switch ( dir ) {
		case 0:
		  pixval=imageout.getPixel(x+i,y+j); // centre pixel
		  break;
		case 1:
		  pixval=imageout.getPixel(x+i,y+j) + 1; // horizontal and vertical pixels
		  break;
		case 2:
		  pixval=imageout.getPixel(x+i,y+j) + sqrt2; // diagonal pixels
		  break;
		default:
		  pixval = 0;
		  assert( false );
		};

		if (pixval<minval) minval=pixval;

	      }

	  imageout.pixel(i,j) = minval;
	}


  }


  // Four connected flood fill algorithm
  void mm_binmorph_tool::floodFill(mm_image<int> &imagein, int x, int y, int newval, int oldval)
  {
    if ((x<0) || (x>=imagein.getWidth()) || (y<0) || (y>=imagein.getHeight())) return;
    if (imagein.pixel(x,y)==newval) return;

    imagein.pixel(x,y) = newval;
    floodFill(imagein,x+1,y,newval,oldval);
    floodFill(imagein,x-1,y,newval,oldval);
    floodFill(imagein,x,y+1,newval,oldval);
    floodFill(imagein,x,y-1,newval,oldval);
  }

}
