//
// Method to find and track pipette tip using the Haynes and Jain edge-change
// detector and a recursive search algorithm. Note that this implementation
// assumes that the pipette is approaching from the top of the screen.
// 
// In order for initial recognition to take place the pipette has to be moved
// (ideally horizontally) so that the initial estimate of the tip can be
// obtained.
//
// Bala Amavasai (bala@amavasai.org)
// $Header: /cvs/mimas/lib/mm_findpipette.cc,v 1.11 2005/07/15 19:40:05 engjw Exp $
//

#include "mm_findpipette.h"
#include "mm_image_funcs.h"

namespace mimas {
using namespace std;
   
void mm_findPipette::setSkipXboundaries(int tempval)
{
    skipXboundaries=tempval;
}

void mm_findPipette::setSkipYboundaries(int tempval)
{
    skipYboundaries=tempval;
}

int mm_findPipette::getSkipYboundaries(void)
{
    return skipYboundaries;
}

void mm_findPipette::setYSearchRange(int tempval)
{
    ySearchRange=tempval;
}

void mm_findPipette::setFirstRunFlag(bool tempval)
{
    firstrecog=tempval;
    if (firstrecog==true) scanliney=skipYboundaries;
}

void mm_findPipette::setHJThresholdVal(int tempval)
{
    thresval=tempval;
}

void mm_findPipette::setLinesDiffRange(int tempval)
{
    linesDiffRange=tempval;
}

void mm_findPipette::setHoldPositionCondition(int tempval)
{
    holdPositionCondition=tempval;
}


// recursive function to search downwards for end of line tip
void mm_findPipette::searchDown(int x, int y, int colour, mm_image<int> &imagein, mm_Cartesian<int> &d)
{
	if ((x<0) || (y<0))
	{
		d.x=-1; d.y=-1;
		return;
	}

	if ((imagein.getPixel(x,y)!=colour) && (imagein.getPixel(x,y)!=0))
	{
		imagein.setPixel(x,y,colour);
		if (y>d.y)
		{
			d.y=y;
			d.x=x;
		}
		searchDown(x+1,y,colour,imagein,d);
		searchDown(x-1,y,colour,imagein,d);
		searchDown(x,y+1,colour,imagein,d);

	}
}





void mm_findPipette::find(mm_image<int> &imageprev, mm_image<int> &imagecurrent, int &x, int &y)
{
    mm_image<int> hj( edgeHaynesJain( imageprev, imagecurrent ) );
    normaliseIt( hj, 0, 255 );
    bilevelIt( hj, thresval, 0, 255 );

    mm_Cartesian<int> depth[hj.getWidth()];
    int xpos[hj.getWidth()];

    // xpos is just the x positions - we initialise this to the
    // correct values before it is sorted. This is then sorted
    // according to depth[] below to obtain an estimate of the
    // midpoint of the pipette tip
    for (int i=0; i<hj.getWidth(); i++)
	    xpos[i]=i;

    // The '1000' below is to make the getPixel(x,y) line in
    // searchDown !=colour
    for (int i=skipXboundaries; i<hj.getWidth()-skipXboundaries; i++)
	    searchDown(i,scanliney,1000+i,hj,depth[i]);

    // bubblesort points so that points that are deepest into the
    // scene appears at the the top of depth[] 
    for (int i=0; i<hj.getWidth()-1; i++)
    {
	    for (int j=0; j<hj.getWidth()-1-i; j++)
		    if (depth[j+1].y > depth[j].y)
		    {
			    int tmpxpos, tmpx, tmpy;
			    tmpx=depth[j].x;
			    tmpy=depth[j].y;
			    tmpxpos=xpos[j];

			    depth[j].x=depth[j+1].x;
			    depth[j].y=depth[j+1].y;
			    xpos[j]=xpos[j+1];

			    depth[j+1].x=tmpx;
			    depth[j+1].y=tmpy;
			    xpos[j+1]=tmpxpos;
		    }
    }

    //	for (int i=0; i<hj.getWidth(); i++)
    //	    cout << xpos[i] << "  " << depth[i] << endl;


    // find x coord of pipette tip since there are two or more lines forming the pippette
    // for the y coordinate choose the longest one only which is depth[0].y
    currdepth.x=depth[0].x;
    currdepth.y=depth[0].y;
    int i;
    for (i=1; i<hj.getWidth(); i++)
    {
	    // consider the depths that are at most linesDiffRange pixels
	// different from deepest (depth[0]) and accummulate the x
	// direction - this smoothes away any noise
	    if (currdepth.y-depth[i].y<linesDiffRange)
	    {
		    currdepth.x=currdepth.x + depth[i].x;
	    }
	    else
		    break;
    } 
    // average the x direction which should give the center point
    // fo the pipette in the x direction
    currdepth.x=currdepth.x/i;

    if (!firstrecog)
    {
//	    if ((abs(currdepth.y-prevdepth.y)>holdPositionCondition)
//	        || (abs(currdepth.x-prevdepth.x)>holdPositionCondition))
	    if (abs(currdepth.y-prevdepth.y)>holdPositionCondition)
	    {
		    currdepth.x=prevdepth.x;
		    currdepth.y=prevdepth.y;
	    }
	    else
	    {
		prevdepth.x=currdepth.x;
		prevdepth.y=currdepth.y;
	    }
    }
    else
    {
	    //store initial values
	    prevdepth.x=currdepth.x;
	    prevdepth.y=currdepth.y;
	    firstrecog=false;
    }


    x=currdepth.x;
    y=currdepth.y;


    // In the next loop search down from -ySearchRange pixels from the
    // current estimated pipette tip
    scanliney=depth[0].y-ySearchRange;
    if (scanliney<skipYboundaries) scanliney=skipYboundaries;

}

}
