#include "mm_image_firewireinput.h"

#ifdef HAVE_LIBDC1394
#include "mm_rgba.h"

namespace mimas {

static inline unsigned char limit( int x ) {
  return x > 0xffffff ? 0xff : ( x <= 0xffff ? 0x0 : x >> 16 );
};

void uyvy422_to_rgba( const char *in, int width, int height, char *out )
{
  assert( out != NULL );
  assert( width % 2 == 0 );
  assert( height % 2 == 0 );
#ifndef NDEBUG
  std::cerr << "UYVY422 to RGBA for " << width << "x" << height << "-image."
            << std::endl;
#endif

  const int rvScale =   91881;
  const int guScale =  -22553;
  const int gvScale =  -46801;
  const int buScale =  116129;
  
  // Direct access to image-memory!

  for ( int y=0; y<height; y++ ) {
	  
    for ( int x=0; x<width; x+=2 ) {

      const unsigned char
        *in0 = (const unsigned char *)in + 1,
        *in1 = (const unsigned char *)in + 3;

      unsigned char
        *out0 = (unsigned char *)out,
        *out1 = (unsigned char *)out + 4;

      unsigned int
        in0int = ((unsigned int)*in0) << 16,
        in1int = ((unsigned int)*in1) << 16;

      signed char
        inu = ((signed char *)in)[0] ^ 0x80,
        inv = ((signed char *)in)[2] ^ 0x80;

      int
        rint =                 rvScale * inv,
        gint = guScale * inu + gvScale * inv,
        bint = buScale * inu;

      out0[0] = limit( bint + in0int );
      out0[1] = limit( gint + in0int );
      out0[2] = limit( rint + in0int );
      out0[3] = 0;

      out1[0] = limit( bint + in1int );
      out1[1] = limit( gint + in1int );
      out1[2] = limit( rint + in1int );
      out1[3] = 0;

      in += 4;
      out += 8;
    };
  };
}

template<>
void mm_image_firewireinput< unsigned char >::
read( mm_image< unsigned char > &image ) throw (mm_exception)
{
  int width, height, size;

  // Perform image capture.
  MMERROR( dc1394_dma_single_capture( &camera ) == DC1394_SUCCESS,
           mm_exception, ,
           "Failed capturing frame from firewire digital camera." );

  switch ( mode ) {
  case MODE_640x480_MONO:
    width = 640;
    height = 480;
    size = 640 * 480;
    break;
  default:
    MMERROR( false, mm_exception, ,
             "Video-mode number " << mode << " not supported "
             "by mimas. Please contact the developers." );
  };

  image.init( width, height );
  memcpy( image.rawData().data(), camera.capture_buffer, size );

  // Allow buffer to be reused.
  dc1394_dma_done_with_buffer( &camera );
}

template<>
void mm_image_firewireinput< mm_rgba< unsigned char > >::
read( mm_image< mm_rgba< unsigned char > > &image ) throw (mm_exception)
{
  int width, height;

  // Perform image capture.
  MMERROR( dc1394_dma_single_capture( &camera ) == DC1394_SUCCESS,
           mm_exception, ,
           "Failed capturing frame from firewire digital camera." );

  switch ( mode ) {
  case MODE_320x240_YUV422:
    width = 320;
    height = 240;
    break;
  case MODE_640x480_YUV422:
    width = 640;
    height = 480;
    break;
  default:
    MMERROR( false, mm_exception, ,
             "Video-mode number " << mode << " not supported "
             "by mimas. Please contact the developers." );
  };

  image.init( width, height );
  uyvy422_to_rgba( (const char *)camera.capture_buffer, width, height,
                   (char *)image.rawData().data() );

  // Allow buffer to be reused.
  dc1394_dma_done_with_buffer( &camera );
}

};

#endif
