//
// new pairwise code, ported in part from TINA
// by stu meikle Wed Mar  8 09:35:01 2000
//

/**
 * Pairwise geometric histograms. see the many papers by Evans,Thacker,Ashbrook and Caparrelli
 * This flaky implementation by stuart meikle 22.02.2000, based on original Tina tool source code.
 *
 * Each pgh is a HISTOGRAM. It is a representation of a specific line in a model or scene. It
 * contains a representation of the geometric properties of lines surrounding the specific line.
 *
 */

#include <cmath>
#include "mm_linalg.h"
#include "mm_pgh.h"

namespace mimas {
using namespace std;


/**
 * Constructor for the histogram. Initialises defaults and calculates the erf lookup array.
 *
 */
mm_pgh::mm_pgh(void)
{
    //initialise some default parameters
    //(just copying from the tina pgh tool)
    pairs_type = pgh_type_directed;
    dbin_max = 50.0;
    dbin_size = 5.0;
    dist_ramp = 5.0;
    num_abin = 32;
    angle_sigma = M_PI/32.0;

    pair_dist = pair_angle = NULL;
    real_pair_dist = real_pair_angle = NULL;

    //we won't actually initialise until the user has confirmed these
    //because initialistion allocates the output image

    erfarray.init( 81, 4.0, 0.1 ); //taken from pgh_defs.c
}

/**
 * destructor. deletes any allocated arrays.
 * <FIX> doesn't appear to delete the erfarray though, mysteriously
 */
mm_pgh::~mm_pgh(void)
{
    if (real_pair_dist)	delete[] real_pair_dist;
    if (real_pair_angle) delete[] real_pair_angle;
}


void	mm_pgh::describeLine( const mm_line &ref_line, const mm_line_ptr_list &l )
{
    build_normalized_pairwise(ref_line, l);
}


// inserted by fbc, 12.10.2001
/**
 * an alternative 'constructor' used in the case that alternative parameters are desired
 */
void	mm_pgh::initialise(	mm_pgh_type newpairs_type, double newdbin_max, double newdbin_size,
							double newdist_ramp, int newnum_abin, double newangle_sigma)
{
	pairs_type = newpairs_type;
	dbin_max = newdbin_max;
	dbin_size = newdbin_size;
	dist_ramp = newdist_ramp;
	num_abin = newnum_abin;
	angle_sigma = newangle_sigma;

	initialise();
}


/**
 * initialise the histogram
 */
void	mm_pgh::initialise( void )
{
    //----TINACODE BOUNDARY----
    //----(like a wormhole into the past.....)----

    /* set up the dimensions of the pairwise histogram NAT 11/11/92 */
    int ndbin;

    if (real_pair_dist !=NULL)
		delete[] real_pair_dist;//delete real_pair_dist;
    if (real_pair_angle !=NULL)
		delete[] real_pair_angle;//delete real_pair_angle;

    dbin_min = -dbin_max; // why! why! why!
			     // negative distance measurements are not possible are they?
			     // @@@ ###
    if (pairs_type == pgh_type_mirror)
	dbin_min = -dist_ramp;

    abin_size = M_PI/num_abin;

    if (pairs_type == pgh_type_directed)
	num_abin*=2;

    ndbin = (int)((dbin_max - dbin_min)/dbin_size);
    pair_dist = new double[ ndbin ];

    //unbelievably bad practise really. ho hum.
    real_pair_dist = pair_dist;
    pair_dist -= (int)floor(dbin_min/dbin_size);
    pair_angle = new double[ 2*num_abin + 2 * (int)floor(1.0+3.0*angle_sigma/abin_size) ];
    real_pair_angle = pair_angle;
    pair_angle += (int)floor( 1.0+3.0*angle_sigma/abin_size );

    //init_erf();
    // already done

    //NO! , like are you mmmmmaaadddddddd?!?!?!?
    //set_pairs_rotate(dbin_min,dbin_max,dbin_size,num_abin,pair_dist,
    //                 pair_angle,abin_size,angle_sigma,dist_ramp);
    //set_pairs_mirror(dbin_min,dbin_max,dbin_size,num_abin,pair_dist,
    //                 pair_angle,abin_size,angle_sigma,dist_ramp);
    //set_pairs_directed(dbin_min,dbin_max,dbin_size,num_abin,pair_dist,
    //                 pair_angle,abin_size,angle_sigma,dist_ramp);
    //----TINACODE BOUNDARY----
}



/**
 * see Tinatool source. adjust the weight associated with a given distance m'ment, if that
 * distance m'ment is less than the specified dist_ramp
 */
bool    mm_pgh::adjust_for_distance_ramp(
				  double *weight,  double *min_d,
				  double *max_d,double dbin_min,
				  double dist_ramp )
{
    //SUSPICIOUS methodology here i think!
    //if the distance is less than a predetermined distance then
    //set to that predet. distance and weight accordingly.

    //store the new values of weight and min_dist back into the given vars
    //stu Tue Mar  7 14:25:06 2000

   if (*min_d>dbin_max-dist_ramp) return false;
   if (*max_d<dbin_min+dist_ramp) return false;
   if (*min_d<dbin_min+dist_ramp)
   {
      *weight *= (dbin_min+dist_ramp-*min_d)/(*max_d-*min_d);
      *min_d = dbin_min+dist_ramp;
   }

   //also adjuust weight for greater distances
   if (*max_d>dbin_max-dist_ramp)
   {
      *weight *= (dbin_max-dist_ramp-*min_d)/(*max_d-*min_d);
      *max_d = dbin_max-dist_ramp;
   }

   return true;
}




//i'll keep these wonderful function names private ;-)
/**
 * create a pairwise histogram for the ref_line given , encoding the geometrical properties
 * of the other lines (geom) about that ref_line.
 */
void	mm_pgh::build_normalized_pairwise(const mm_line &ref_line, const mm_line_ptr_list &geom )
{

//    double    l, l0;
    int		i;

        /* (1) Allocate space for the histogram.
           (2) Compile the histogram from the ref line and
               the line list.
           (3) Normalize the histogram.
           (4) Sqrt the hist to get the correct statistics.
           (5) Add a Hist_ref to the histograms props list.
        */

    int h = (int)floor(dbin_max/dbin_size) - (int)floor(-dbin_max/dbin_size);
    init( num_abin, h );
    //(always returns a la ANSI C++ standard for 'new')

    i=0;
    for( mm_line_ptr_list::const_iterator	listit = geom.begin();
         listit!=geom.end(); listit++, i++ )
    {
	 	const Vector *linetips;
		mm_line_ptr l=*listit;
		double x1,x2,y1,y2;
		double yl1, xl1, yl2, xl2;

		linetips = &ref_line.getStart();
		x1=(*linetips)(0);
		y1=(*linetips)(1);
		linetips = &ref_line.getEnd();
		x2=(*linetips)(0);
		y2=(*linetips)(1);

		xl1=(x1-x2)/2.0;
		yl1=(y1-y2)/2.0;

		linetips=&l->getStart();
		x1=(*linetips)(0);
		y1=(*linetips)(1);
		linetips=&l->getEnd();
		x2=(*linetips)(0);
		y2=(*linetips)(1);
		xl2=(x1-x2)/2.0;
		yl2=(y1-y2)/2.0;

	 	if ((((xl1-xl2)*(xl1-xl2))+((yl1-yl2)*(yl1-yl2))) <= (dbin_max*dbin_max))
        compare_lines( ref_line, **listit );
    }

//    l0 = ref_line->getStart()->distanceTo( *ref_line->getEnd() );
//    l = sum_pgh(pairwise)/l0;

    sqr_root_and_normalize();

/* --- removed. don't think we need this information now .
// stu Thu 06 Apr 2000 15:28:31 BST
    ref = ralloc(sizeof(Hist_ref));
    if (ref==NULL)
    {
        format("Error: No space for Hist_ref.\n");
        im_free(pairwise);
        return NULL;
    }

    ref->model = ref_model;
    ref->matches = NULL;
    ref->line_seg = ref_line;
    ref->l0 = l0;
    ref->l = l;
    ref->type = type;
    ref->scale_factor = 1.0;

    pairwise->props = proplist_add(pairwise->props, ref, HIST_REF_TYPE,
                                                               hist_ref_free);

    pairwise->label = ref_line->label;
--- */

    //----TINACODE BOUNDARY----
}



/**
 * used during the creation process.
 * add pgh contributions for lines l1 and l2 to histogram in im
 * l1 is reference line, l2 is comparison line.
 */
void 	mm_pgh::compare_lines(const mm_line &refline, const mm_line &line)
/* add pgh contributions for lines l1 and l2 to histogram in im */
/* l1 is reference line, l2 is comparison line.         NAT 29/7/93 */
{
    switch(pairs_type)
    {
      case pgh_type_mirror:	compare_lines_mirror(refline,line);	break;
      case pgh_type_rotate:      compare_lines_rotate(refline,line); 	break;
      case pgh_type_directed:    compare_lines_directed(refline,line); 	break;
    }
}



//---------------------- MIRROR ----------------------------------------
/**
 * add pairwise contributions for lines l1 and l2 to histogram in im
 * l1 is reference line, l2 is comparison line.
 */
void 	mm_pgh::compare_lines_mirror(const mm_line &l1, const mm_line &l2)
// add pairwise contributions for lines l1 and l2 to histogram in im
// l1 is reference line, l2 is comparison line.         NAT 10/11/92
{

  Vector v1(2),v2(2),normv1(2),isct(2);
    bool	l1_intsct,l2_intsct;
    double 	weight,min_d,max_d,angle; //!!! beware name conflict with instance vs

    isct = l1.getIntersection(l2,&l1_intsct,&l2_intsct);

    if( !l1_intsct && !l2_intsct)
    {
	// both lines no intersections
       v1 = dir_vec(isct,l1.getStart(),l1.getEnd());
       v2 = dir_vec(isct,l2.getStart(),l2.getEnd());
       angle = fabs( mimas::angle( v1, v2 ) );
       Vector unitv( unit( l1.getVector() ) );
       normv1( 0 ) = -unitv(1); normv1( 1 ) = unitv(0);

       min_d = fabs( inner_prod( normv1, l2.getStart() - isct ) );
       max_d = fabs( inner_prod( normv1, l2.getEnd()   - isct ) );
       weight   = l1.length() * l2.length();

       make_entry_mirror( min_d, max_d, angle, weight );

       //originals
       //min_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p1,isct)));
       //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p2,isct)));
       //weight = vec2_dist(l1.p1,l1.p2)*vec2_dist(l2.p1,l2.p2);
       //make_entry_mirror(im,min_dist,max_dist,angle,weight);

    }
    else
    {
       if(l2_intsct)
       {
          if(l1_intsct)
          {
	    // l1_p1 to isct and l2_p1 to isct
            Vector unitv( unit( l1.getVector() ) );
            v1 = dir_vec(isct,l1.getStart(),isct);
            v2 = dir_vec(isct,l2.getStart(),isct);
       angle = fabs(mimas::angle(v1,v2));
	    normv1(0) = -unitv(1); normv1(1) = unitv(0);
	    min_d = 0.0;
	    max_d = fabs( inner_prod( normv1, l2.getStart() - isct ) );
	    weight= norm_2( isct - l1.getStart() ) * norm_2( isct - l2.getStart() );

	    make_entry_mirror( min_d, max_d, angle, weight );

	    //original - left in place for debugging purposes!
            //angle = fabs(vec2_angle(v1,v2));
            //normv1 = vec2(-((l1.v).el[1]),(l1.v).el[0]);
            //min_dist = 0.0;
            //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p1,isct)));
            //weight = vec2_dist(l1.p1,isct)*vec2_dist(l2.p1,isct);
            //make_entry_mirror(im,min_dist,max_dist,angle,weight);

	    // isct to l1_p2 and l2_p1 to isct
	    weight = norm_2( isct - l1.getEnd() ) * norm_2( isct - l2.getStart() );
	    make_entry_mirror(min_d,max_d,M_PI-angle,weight);

            //weight = vec2_dist(l1.p2,isct)*vec2_dist(l2.p1,isct);
            //make_entry_mirror(im,min_dist,max_dist,M_PI-angle,weight);

	    // l1-p1 to isct and isct to l2_p2
	    max_d  = fabs( inner_prod( normv1, l2.getEnd() - isct ) );
	    weight = norm_2( isct - l1.getStart() ) * norm_2( l2.getEnd() - isct );
	    make_entry_mirror( min_d, max_d, M_PI-angle, weight );

    	    //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p2,isct)));
            //weight = vec2_dist(l1.p1,isct)*vec2_dist(isct,l2.p2);
            //make_entry_mirror(im,min_dist,max_dist,M_PI-angle,weight);

	    // isct to l1_p2 and isct to l2_p2
	    weight = norm_2( isct - l1.getEnd() ) * norm_2( l2.getEnd() - isct );
	    make_entry_mirror( min_d, max_d, angle , weight );

            //weight = vec2_dist(l1.p2,isct)*vec2_dist(isct,l2.p2);
            //make_entry_mirror(im,min_dist,max_dist,angle,weight);
	  }
          else
          {
	     // l1_p1 to l1_p2 and l2_p1 to isct
            Vector unitv( l1.getVector() );

             v1 = dir_vec(isct,l1.getStart(),l1.getEnd());
             v2 = dir_vec(isct,l2.getStart(),isct);

        angle  = fabs( mimas::angle( v1, v2 ));
        normv1( 0 ) = -unitv(1); normv1( 1 ) = unitv(0);
	     min_d  = 0.0;
	     max_d  = fabs( inner_prod( normv1, l2.getStart() - isct ) );
	     weight = l1.length() * norm_2( isct - l2.getStart() );
	     make_entry_mirror( min_d, max_d, angle, weight );

	     // angle = fabs(vec2_angle(v1,v2));
             //normv1 = vec2(-((l1.v).el[1]),(l1.v).el[0]);
             //min_dist = 0.0;
             //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p1,isct)));
             //weight = vec2_dist(l1.p1,l1.p2)*vec2_dist(l2.p1,isct);
             //make_entry_mirror(im,min_dist,max_dist,angle,weight);

	     // l1_p1 to l1_p2 and isct to l2_p2
	     max_d  = fabs( inner_prod( normv1, l2.getEnd() - isct ) );
	     weight = l1.length() * norm_2( isct - l2.getEnd() );
	     make_entry_mirror( min_d, max_d, M_PI-angle, weight );

	     //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p2,isct)));
             //weight = vec2_dist(l1.p1,l1.p2)*vec2_dist(isct,l2.p2);
             //make_entry_mirror(im,min_dist,max_dist,M_PI-angle,weight);
          }
       }
       else
       {
	   // l1_p1 to isct and l2_p1 to l2_p2
         Vector unitv( l1.getVector() );

           v1 = dir_vec(isct,l1.getStart(),isct);
           v2 = dir_vec(isct,l2.getStart(),l2.getEnd());

	   angle  = fabs( mimas::angle( v1, v2 ) );
      normv1(0) = -unitv(1); normv1(1) = unitv(0);
	   min_d  = fabs( inner_prod( normv1, l2.getStart() - isct ));
	   max_d  = fabs( inner_prod( normv1, l2.getEnd()   - isct ));
	   weight = norm_2( isct - l1.getStart() ) * l2.length();
	   make_entry_mirror( min_d, max_d, angle, weight );

	   //angle = fabs(vec2_angle(v1,v2));
           //normv1 = vec2(-((l1.v).el[1]),(l1.v).el[0]);
           //min_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p1,isct)));
           //max_dist = fabs(vec2_dot(normv1,vec2_diff(l2.p2,isct)));
           //weight = vec2_dist(l1.p1,isct)*vec2_dist(l2.p1,l2.p2);
           //make_entry_mirror(im,min_dist,max_dist,angle,weight);

	   // isct to l1_p2 and l2_p1 to l2_p2
	   weight = norm_2( isct - l1.getEnd() ) * l2.length();
	   make_entry_mirror( min_d, max_d, M_PI-angle, weight );

           //weight = vec2_dist(isct,l1.p2)*vec2_dist(l2.p1,l2.p2);
           //make_entry_mirror(im,min_dist,max_dist,M_PI-angle,weight);
       }
    }
}

void mm_pgh::make_entry_mirror(double min_d,double max_d,double angle,double weight)
{
   int 	  low_angle_bin,high_angle_bin;
   int 	  low_dist_bin,high_dist_bin;
   int 	  bin_edge,dbin,gbin;
   int    angle_bin;
   double current_contents;
   double angle_fac;
   double temp;

    if (!initialised()) //mm image call
	return;

   if (min_d>max_d)
   {
       temp  = min_d;
       min_d = max_d;
       max_d = temp;
   }

   if (!adjust_for_distance_ramp( &weight, &min_d, &max_d, dbin_min, dist_ramp ))
       return;

   get_pdist(pair_dist,min_d,max_d,dbin_size,dist_ramp,
              &low_dist_bin,&high_dist_bin);
   get_pdist(pair_angle,angle-angle_sigma,angle+angle_sigma,abin_size,
              2.0*angle_sigma,&low_angle_bin,&high_angle_bin);

   for (bin_edge = low_angle_bin;bin_edge<=high_angle_bin; bin_edge++ )
   {
      angle_fac = pair_angle[bin_edge];
      for (dbin = low_dist_bin;dbin<=high_dist_bin;dbin++)
      {
          angle_bin = bin_edge;
          while (angle_bin<0) angle_bin = angle_bin+num_abin;
          while (angle_bin>=num_abin) angle_bin = angle_bin-num_abin;
          gbin = dbin;
 
          if (dbin<0)
          {
            gbin = - dbin -1;
            angle_bin = num_abin - angle_bin -1;
            if (angle_bin>=num_abin) angle_bin -= num_abin;
          }
	  
	  current_contents = getPixel( angle_bin, gbin );
	  setPixel( angle_bin, gbin, current_contents +
                              angle_fac*weight*pair_dist[dbin]    );

          //IM_PIX_GET(im, gbin, angle_bin, current_contents);
          //IM_PIX_SET(im, gbin, angle_bin, current_contents +
          //                              angle_fac*weight*pair_dist[dbin]);

     MMERROR( current_contents + angle_fac*weight*pair_dist[dbin]>=0.0,
              mm_exception, ,
              "nat says -bollocks-" );//artistic license
      }
   }
}

//---------------------- ROTATE -----------------------------------------
/**
 * add pairwise contributions for lines l1 and l2 to histogram in im
 * l1 is reference line, l2 is comparison line.
 */

void 	mm_pgh::make_entry_rotate(
			double min_d, double max_d,
                        double angle, double weight)
{
   int low_angle_bin,high_angle_bin;
   int low_dist_bin,high_dist_bin;
   int bin_edge,dbin,gbin;
   int angle_bin;
   double current_contents;
   double angle_fac;
   double temp;

   if (!initialised())
	return;

   if (min_d>max_d)
   {
       temp = min_d;
       min_d = max_d;
       max_d = temp;
   }

   if (!adjust_for_distance_ramp( &weight, &min_d, &max_d, dbin_min, dist_ramp ))
       return;

   get_pdist(pair_dist,min_d,max_d,dbin_size,dist_ramp,
             &low_dist_bin,&high_dist_bin);
   get_pdist(pair_angle,angle-angle_sigma,angle+angle_sigma,abin_size,
           2.0*angle_sigma,&low_angle_bin,&high_angle_bin);

/* gaussian blurring version NAT 11/11/93
   get_gdist(pair_angle,angle,abin_size,angle_sigma,
                      &low_angle_bin,&high_angle_bin);
*/

   for (bin_edge = low_angle_bin;bin_edge<=high_angle_bin; bin_edge++ )
   {
      angle_fac = pair_angle[bin_edge];
      for (dbin = low_dist_bin;dbin<=high_dist_bin;dbin++)
      {
          if (bin_edge<0)
          {
               angle_bin = bin_edge+num_abin;
               gbin = -dbin -1;
          }
          else if (bin_edge>=num_abin)
          {
               angle_bin = bin_edge-num_abin;
               gbin = -dbin -1;
          }
          else /* generally the case */
          {
               angle_bin = bin_edge;
               gbin = dbin;
          }
 
	  current_contents = getPixel( angle_bin, gbin );
	  setPixel( angle_bin, gbin, current_contents + 
                                          angle_fac*weight*pair_dist[dbin]);			    

          //IM_PIX_GET(im, gbin, angle_bin, current_contents);
          //IM_PIX_SET(im, gbin, angle_bin, current_contents +
          //                                  angle_fac*weight*pair_dist[dbin]);
	  
     MMERROR( current_contents + angle_fac*weight*pair_dist[dbin]>=0.0,
              mm_exception, ,
              "nat says -bollocks-. negative entries in pairwise" );
      }
   }
}



void	mm_pgh::compare_lines_rotate(const mm_line &l1, const mm_line &l2)
/* add pairwise contributions for lines l1 and l2 to histogram in im */
/* l1 is reference line, l2 is comparison line.         NAT 10/11/92 */
{
    Vector v1(2),normv1(2),isct(2);
    bool	l1_intsct, l2_intsct;
    double	weight, min_d, max_d, angle;

    isct = l1.getIntersection(l2,&l1_intsct,&l2_intsct);

    if( !l1_intsct)
    {
	// both lines no intersections 
	v1 = dir_vec(isct,l1.getStart(),l1.getEnd());
	angle = mimas::angle( l2.getVector(), v1 );
	if (angle<0.0) angle += M_PI;

	normv1(0) = -v1(1); normv1(1) = v1(0);
	min_d  = inner_prod( normv1, l2.getStart() - isct );
	max_d  = inner_prod( normv1, l2.getEnd()   - isct );
	weight = l1.length() * l2.length();
	make_entry_rotate( min_d, max_d, angle, weight );
	
	// originals 
	//angle = vec2_angle(l2.getVector(),v1);
	//if (angle<0.0) angle += M_PI;
	//normv1 = vec2(-v1.el[1],v1.el[0]);
        //min_dist = vec2_dot(normv1,vec2_diff(l2.p1,isct));
        //max_dist = vec2_dot(normv1,vec2_diff(l2.p2,isct));
        //weight = vec2_dist(l1.p1,l1.p2)*vec2_dist(l2.p1,l2.p2);
        //make_entry_rotate(im,min_dist,max_dist,angle,weight);
   }
   else
   {
	// l1_p1 to isct and l2_p1 to l2_p2
	v1 = dir_vec( isct, l1.getStart(), isct );
	angle = mimas::angle( l2.getVector(), v1 );
	if (angle<0.0) angle+= M_PI;
	
	normv1(0) = -v1(1); normv1(1) = v1(0);
	min_d  = inner_prod( normv1, l2.getStart() - isct );
	max_d  = inner_prod( normv1, l2.getEnd()   - isct );
	weight = norm_2( isct - l1.getStart() ) * l2.length();
	make_entry_rotate( min_d, max_d, angle, weight );
	
	//orignals
        //v1 = dir_vec(isct,l1.p1,isct);
        //angle = vec2_angle(l2.v,v1);
        //if (angle<0.0) angle += M_PI; 
        //normv1 = vec2(-v1.el[1],v1.e l[0]);
        //min_dist = vec2_dot(normv1,vec2_diff(l2.p1,isct));
        //max_dist = vec2_dot(normv1,vec2_diff(l2.p2,isct));
        //weight = vec2_dist(l1.p1,isct)*vec2_dist(l2.p1,l2.p2);
        //make_entry_rotate(im,min_dist,max_dist,angle,weight);
 
	// isct to l1_p2 and l2_p1 to l2_p2 
	weight = norm_2( l1.getEnd() - isct ) * l2.length();
	make_entry_rotate( -min_d, -max_d, angle, weight );
	
	//originals:
	//weight = vec2_dist(isct,l1.p2)*vec2_dist(l2.p1,l2.p2);
        //make_entry_rotate(im,-min_dist,-max_dist,angle,weight);
   }
}

//---------------------- DIRECTED ---------------------------------------
/**
 * add pairwise contributions for lines l1 and l2 to histogram in im
 * l1 is reference line, l2 is comparison line.
 */

void 	mm_pgh::make_entry_directed(
			 double min_d, double max_d,
                         double angle, double weight)
/* treat the poles in the representation at 0*M_PI and 2*M_PI */
{
   double adiff;

   if ((adiff = fabs(angle))<angle_sigma)
   {
      make_entry_direct(min_d,max_d,angle,
                        weight*(0.5+0.5*adiff/angle_sigma));
      make_entry_direct(min_d,max_d,angle+M_PI,
                        weight*(0.5-0.5*adiff/angle_sigma));
   }
   else if ((adiff = fabs(angle-M_PI))<angle_sigma)
   {
      make_entry_direct(min_d,max_d,angle,
                        weight*(0.5+0.5*adiff/angle_sigma));
      if ((angle-M_PI)>=0.0)
          make_entry_direct(min_d,max_d,angle-M_PI,
                               weight*(0.5-0.5*adiff/angle_sigma));
      else
          make_entry_direct(min_d,max_d,angle+M_PI,
                               weight*(0.5-0.5*adiff/angle_sigma));
   }
   else if ((adiff=fabs(angle-2.0*M_PI))<angle_sigma)
   {
      make_entry_direct(min_d,max_d,angle,
                           weight*(0.5+0.5*adiff/angle_sigma));
      make_entry_direct(min_d,max_d,angle-M_PI,
                           weight*(0.5-0.5*adiff/angle_sigma));

   }
   else  make_entry_direct(min_d,max_d,angle,weight);
}

void mm_pgh::make_entry_direct(double min_d, double max_d,
                       double angle, double weight)
{
   int	  low_angle_bin,high_angle_bin;
   int    low_dist_bin,high_dist_bin;
   int 	  bin_edge,dbin,gbin;
   int 	  angle_bin;
   double current_contents;
   double angle_fac;
   double temp;


   if (!initialised())
	return;

   if (min_d>max_d)
   {
       temp = min_d;
       min_d = max_d;
       max_d = temp;
   }

    if (!adjust_for_distance_ramp( &weight, &min_d, &max_d, dbin_min, dist_ramp ))
	return;

   get_pdist(pair_dist,min_d,max_d,dbin_size,dist_ramp,
                                         &low_dist_bin,&high_dist_bin);
   get_pdist(pair_angle,angle-angle_sigma,angle+angle_sigma,abin_size,
                   2.0*angle_sigma,&low_angle_bin,&high_angle_bin);

   for (bin_edge = low_angle_bin;bin_edge<=high_angle_bin; bin_edge++ )
   {
      angle_fac = pair_angle[bin_edge];
      for (dbin = low_dist_bin;dbin<=high_dist_bin;dbin++)
      {
          angle_bin = bin_edge;
          while (angle_bin<0) angle_bin = angle_bin+num_abin;
          while (angle_bin>=num_abin) angle_bin = angle_bin-num_abin;
          gbin = dbin;

	  current_contents = getPixel( angle_bin, gbin );
	  setPixel( angle_bin, gbin, current_contents +
                                          angle_fac*weight*pair_dist[dbin]);


     MMERROR( (current_contents + (angle_fac*weight*pair_dist[dbin]))>=0.0,
              mm_exception, , "nat says -bollocks-" );

      }
   }
}

void mm_pgh::compare_lines_directed(const mm_line &l1, const mm_line &l2)
/* add pairwise contributions for lines l1 and l2 to histogram in im */
/* l1 is reference line, l2 is comparison line.         NAT 10/11/92 */
{
    Vector v1(2), normv1(2), isct(2);
    bool	l1_intsct,l2_intsct;
    double	weight, min_d,max_d, angle;


    isct = l1.getIntersection(l2,&l1_intsct,&l2_intsct);

    if(!l1_intsct)
    {
      Vector unitv( unit( l1.getVector() ) );

	// both lines no intersections
	v1 = dir_vec(isct,l1.getStart(),l1.getEnd());
	angle = mimas::angle( l2.getVector(), v1 );
        if (angle<0.0) angle += M_PI;

	// if intersection direction disagrees with nominal line direction 
	// then flip into Pi-2Pi domain of plot
   if ( inner_prod( unitv, v1 ) < 0.0)
	    angle += M_PI;
	
	normv1(0) = -unitv(1); normv1(1) = unitv(0);
	min_d = inner_prod( normv1, l2.getStart() - isct );
	max_d = inner_prod( normv1, l2.getEnd()   - isct );
	weight= l1.length() * l2.length();
		
	make_entry_directed( min_d, max_d, angle, weight );
	
	// originals
        //if (vec2_dot(l1.v,v1)<0.0) angle += M_PI;
        //normv1 = vec2(-((l1.v).el[1]),(l1.v).el[0]);
        //min_dist = vec2_dot(normv1,vec2_diff(l2.p1,isct));
        //max_dist = vec2_dot(normv1,vec2_diff(l2.p2,isct));
        //weight = vec2_dist(l1.p1,l1.p2)*vec2_dist(l2.p1,l2.p2);
        //make_entry_directed(im,min_dist,max_dist,angle,weight);l396
	
   }
   else
   {
     Vector unitv( unit( l1.getVector() ) );
   
	// l1_p1 to isct and l2_p1 to l2_p2 
   v1 = dir_vec(isct,l1.getStart(),isct);
   angle = mimas::angle( l2.getVector(), v1 );
   if (angle<0.0) angle += M_PI;
	// if intersection direction disagrees with nominal line direction
	// then flip into Pi-2Pi domain of plot
   if ( inner_prod( unitv,  v1 ) < 0.0)
	    angle += M_PI;

	normv1(0) = -unitv(1); normv1(1) = unitv(0);
	min_d  = inner_prod( normv1, l2.getStart() - isct );
	max_d  = inner_prod( normv1, l2.getEnd()   - isct );
	weight = norm_2( isct - l1.getStart() ) * l2.length();
	make_entry_directed( min_d, max_d, angle, weight );

	//originals
        //normv1 = vec2(-((l1.v).el[1]),(l1.v).el[0]);
        //min_dist = vec2_dot(normv1,vec2_diff(l2.p1,isct));
        //max_dist = vec2_dot(normv1,vec2_diff(l2.p2,isct));
        //weight = vec2_dist(l1.p1,isct)*vec2_dist(l2.p1,l2.p2);
        //make_entry_directed(im,min_dist,max_dist,angle,weight);
 
	// isct to l1_p2 and l2_p1 to l2_p2
	weight = norm_2( isct - l1.getEnd() ) * l2.length();
	make_entry_directed( min_d, max_d, M_PI+angle, weight );
	
	//originals
        //weight = vec2_dist(isct,l1.p2)*vec2_dist(l2.p1,l2.p2);
        //make_entry_directed(im,min_dist,max_dist,M_PI+angle,weight);
   }
}



// supplementary private functions --------------------------------------

/**
 * routine to calculate the direction of the line p1_p2 away from point of intersection.
 */
mm_pgh::Vector mm_pgh::dir_vec( const Vector &isct,
                                const Vector &p1, const Vector &p2 )
/* routine to calculate the direction of the line p1_p2 away from isct */
/* replacement for Aluns modify dir_vec.                  NAT 11/11/92 */
{
    Vector tv(2);

    if( norm_2( isct - p1 ) > norm_2( isct - p2 ) )
    {
      // 'tv = unit( p1 - p2 )' won't work in release-mode for some reason.
      Vector diff( p1 - p2 );
      tv = unit( diff );
    }
    else
    {
      // 'tv = unit( p2 - p1 )' won't work in release-mode for some reason.
      Vector diff( p2 - p1 );
      tv = unit( diff );
    }


  return (tv);
}




/**
 * routine to fill a perpedicular distance array with ramped weighted values
 */
void mm_pgh::get_pdist(double *pdist,double pdmin,double pdmax,
			    double bin_width,double ramp,
			    int *low_bin,int *high_bin)
/* routine to fill a perpedicular distance array with ramped weighted values */
/*                                                              NAT 27/10/92 */
{
    //we don't need to pass all the vars as parameters now, but i'll leave
    //it in for easing the amount of debugging i'll have to do ! stu.

    double pbmin,pbmax;
    double pddiff,ptmin,ptmax,mean;
    double total=0.0;
    int i,j;

    pddiff = pdmax - pdmin;
    mean =  (pdmax+pdmin)/2.0;
    pbmin = mean - 0.5*pddiff - 0.5*ramp;
    pbmax = mean + 0.5*pddiff + 0.5*ramp;
    for (i = (int)floor(pbmin/bin_width) ; i <=(int)floor(pbmax/bin_width); i++)
       pdist[i]=0.0;

    // set the blurring width of the data to 'ramp' or 'pddiff' whichever is smaller
    if (pddiff<ramp)
    {
        ptmin = pbmin + pddiff;
        ptmax = pbmax - pddiff;
    }
    else
    {
        ptmin = pbmin + ramp;
        ptmax = pbmax - ramp;
    }

    // deal with the linear section of the trapezium
    i = (int)floor(ptmin/bin_width);
    j = (int)floor(ptmax/bin_width);
    if (i==j)
    {
        pdist[i] = (double)((ptmax-ptmin)/bin_width);
    }
    else
    {
        pdist[i] = (double)(1.0+(double)i-ptmin/bin_width);
        for (i = (int)floor(ptmin/bin_width) +1 ; i<(int)floor(ptmax/bin_width); i++)
            pdist[i] = (double)1.0;
        pdist[j] = (double)(ptmax/bin_width-(double)j);
    }

    //  now fill leading and trailing triangles

    i = (int)floor(pbmin/bin_width);
    j = (int)floor(ptmin/bin_width);
    if (i==j)
    {
        pdist[i] += (double)(0.5*(ptmin-pbmin)/bin_width);
    }
    else
    {
        pdist[i] += (double)(0.5*(1.0+(double)i-pbmin/bin_width)*(((double)i+1.0)*bin_width-pbmin)/(ptmin-pbmin));
        for (i = (int)floor(pbmin/bin_width) +1 ; i<(int)floor(ptmin/bin_width); i++)
            pdist[i] += (double)((((double)i+0.5)*bin_width-pbmin)/(ptmin-pbmin));
        pdist[j] += (double)((((double)j*bin_width+ptmin)/2.0 - pbmin)*(ptmin/bin_width-(double)j)/(ptmin-pbmin));
    }

    i = (int)floor(ptmax/bin_width);
    j = (int)floor(pbmax/bin_width);

    if (i==j)
    {
        pdist[i] += (double)(0.5*(pbmax-ptmax)/bin_width);
    }
    else
    {
        pdist[i] += (double)((pbmax - (((double)i+1)*bin_width+ptmax)/2.0)
                   *(((double)i+1.0)-ptmax/bin_width)/(pbmax-ptmax));

        for (i = (int)floor(ptmax/bin_width) +1 ; i<(int)floor(pbmax/bin_width); i++)
            pdist[i] += (double)((pbmax-((double)i+0.5)*bin_width)/(pbmax-ptmax));
        pdist[j] += (double)(0.5*(pbmax-(double)j*bin_width)*(pbmax/bin_width-(double)j)/(pbmax-ptmax));
    }

    // now normalise

   for (i = (int)floor(pbmin/bin_width) ; i <=(int)floor(pbmax/bin_width); i++)
       total += pdist[i];
   for (i = (int)floor(pbmin/bin_width) ; i <=(int)floor(pbmax/bin_width); i++)
       pdist[i] /= (double)total;
   *low_bin = (int)floor(pbmin/bin_width);
   *high_bin = (int)floor(pbmax/bin_width);

}


/**
 * post creation processing of the histogram. square root and normalise the histogram
 * really this is normalise and then squareroot!
 */
void	mm_pgh::sqr_root_and_normalize(void)
{
    double	sum;
    int		i, j;

    for(sum = 0, j = 0; j < getHeight(); j++)
		for(i = 0; i < getWidth(); i++)
	    	sum += fabs(pixel(i, j));

    if (sum < 0.0001) return;

    for(j = 0;j < getHeight(); j++)
    {
		for(i = 0; i < getWidth(); i++)
		{
	    	double 	nv;

	    	nv = getPixel(i, j) / sum;
	    	if (nv > 0) nv = sqrt(nv);

	    	pixel(i, j) = nv;
		}
    }
}
}
