#ifndef NDEBUG
#include <iostream>
#endif
#include <X11/keysym.h>
#include "mm_x11_window.h"

using namespace std;

namespace mimas {

mm_x11_display_ptr mm_x11_window::display;

bool mm_x11_window::state = true;

// List of windows.
std::map< Window, std::pair< void (*)(void *), void * > >
  mm_x11_window::windows;

mm_x11_window::mm_x11_window( mm_x11_display_ptr _display,
                              int _width, int _height,
                              void (*_paint)(void *), void *_arg,
                              bool depthBuffer ):
  width(_width), height(_height)
{
  state = true;
  MMERROR( width > 0 && height > 0, mm_exception, ,
           width << 'x' << height << " is an illegal image-size." );
  
  if ( !windows.empty() ) {
    // Get connection to X11 display
    MMERROR( display == _display, mm_exception, ,
             "mm_x11_window only works for a single display." );
  } else
    display = _display;

  try {
  
    // Request desired X11 visual.
    int attribs[] = { GLX_RGBA,
                      GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1,
                      GLX_DEPTH_SIZE, depthBuffer ? 16 : 0, None };
    visualInfo = glXChooseVisual( display->get(),
                                  DefaultScreen( display->get() ),
                                  attribs );
    MMERROR( visualInfo != NULL, mm_exception, ,
             "Error choosing desired X11 visual." );
      
    try {
      
      // Create a color map.
      colorMap = XCreateColormap( display->get(),
                                  RootWindow( display->get(),
                                              visualInfo->screen ),
                                  visualInfo->visual, AllocNone );
      MMERROR( colorMap != 0, mm_exception, ,
               "Couldn't allocate X11 colormap." );
        
      try {
          
        XSetWindowAttributes attributes;
        attributes.colormap = colorMap;
        attributes.border_pixel = 0;
        attributes.event_mask =
          ExposureMask | StructureNotifyMask | KeyPressMask;

#ifndef NDEBUG
        cerr << "Creating window of size " << _width << 'x' << _height << endl;
#endif
          
        window = XCreateWindow( display->get(),
                                RootWindow( display->get(),
                                            visualInfo->screen), 0,
                                0, width, height,
                                0, visualInfo->depth, InputOutput,
                                visualInfo->visual,
                                CWBorderPixel | CWColormap | CWEventMask,
                                &attributes );
          
        MMERROR( window != 0, mm_exception, , "Error creating X11 window." );

        windows[ window ] = make_pair( _paint, _arg );
          
        try {
            
          // Wait for X-server.
          XEvent event;
          XMapWindow( display->get(), window );
          XIfEvent( display->get(), &event, waitForNotify, (char *)window );
        
        } catch ( mm_exception &e ) {

          windows.erase( window );
          XDestroyWindow( display->get(), window );
          throw e;
        
        };
          
      } catch ( mm_exception &e ) {
      
        XFreeColormap( display->get(), colorMap );
        throw e;
      
      };
        
    } catch ( mm_exception &e ) {
    
      XFree( visualInfo );
      throw e;
    
    };

  } catch ( mm_exception &e ) {

    if ( windows.empty() )
      display.reset();
    throw e;

  };

 }

mm_x11_window::~mm_x11_window(void)
{
  XDestroyWindow( display->get(), window );
  XFreeColormap( display->get(), colorMap );
  XFree( visualInfo );

  windows.erase( window );
  if ( windows.empty() )
    display.reset();
}

Bool mm_x11_window::waitForNotify( Display *, XEvent *e, char *arg )
{
  return ( e->type == MapNotify ) && ( e->xmap.window == (Window)arg );
}

void mm_x11_window::process_events( bool wait ) throw (mm_exception)
{
  MMERROR( state, mm_exception, ,
           "User pressed escape after window was opened." );
  XEvent event;
  // XNextEvent( display, &event );
  
  while ( true ) {
    if ( wait )
      XMaskEvent( display->get(),
                  KeyPressMask | ExposureMask | StructureNotifyMask,
                  &event );
    else
      if ( !XCheckMaskEvent( display->get(),
                             KeyPressMask | ExposureMask | StructureNotifyMask,
                             &event ) )
        break;

    if ( event.type == Expose || event.type == ConfigureNotify ) {

      // Look for window.
      map< Window, pair< void (*)(void *), void * > >::iterator entry =
        windows.find( event.xany.window );
      
      if ( entry != windows.end() ) {
        // Invoke paint-routine.
        (*entry->second.first)( entry->second.second );
      };

    } else if ( event.type == KeyPress ) {

#ifndef NDEBUG
      cerr << "keycode = " << event.xkey.keycode << endl;
#endif
      if ( event.xkey.keycode == 0x41 )// space
        break;
      else if ( event.xkey.keycode == 0x09 ) {// escape
        state = false;
        break;
      };
      
    };

  };

}

void mm_x11_window::resize( int _width, int _height )
{
  if ( width != _width || height != _height ) {
#ifndef NDEBUG
    cerr << "Resizing to " << _width << "x" << _height << endl;
#endif
    XResizeWindow( display->get(), window, _width, _height );
    width = _width;
    height = _height;
    XFlush( display->get() );
  };
}

};
